# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import Dict, List, Optional

import attr

from .enums import DeploymentState
from .model import BaseModel
from .utils import str_to_datetime


@attr.define()
class DeploymentModelIdentifier:
    """
    Container class holding the unique identifiers for a model created in an Intel®
    Geti™ project that will be or has been deployed.

    :var model_id: Unique ID of the model to be deployed
    :var model_group_id: Unique ID of the model group to which the model
        belongs
    """

    model_id: str
    model_group_id: str

    def to_dict(self) -> Dict[str, str]:
        """
        Return the dictionary representation of the model identifier.

        :return: Dictionary containing the model identifiers
        """
        return attr.asdict(self)

    @classmethod
    def from_model(cls, model: BaseModel) -> "DeploymentModelIdentifier":
        """
        Get the identifiers needed for model deployment for a given `model`

        :param model: Model to retrieve identifiers for
        """
        return cls(model_id=model.id, model_group_id=model.model_group_id)


@attr.define()
class CodeDeploymentInformation:
    """
    Class containing information pertaining to the deployment of an Intel® Geti™
    project.

    :var id: Unique ID of the code deployment
    :var progress: Progress of the deployment creation process
    :var state: State of the deployment creation process.
    :var models: List of model identifiers for the models involved in the deployment
    :var message: Message providing more details on the state of the deployment
    :var creator_id: String containing the ID of the user who created the deployment
    :var creation_time: Time at which the deployment was created
    """

    id: str
    progress: float
    state: DeploymentState
    models: List[DeploymentModelIdentifier]
    creator_id: str
    creation_time: Optional[str] = attr.field(default=None, converter=str_to_datetime)
    message: str = ""



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from typing import Any, ClassVar, Dict, List, Optional, Sequence, Type, Union, get_args

import attr

from geti_sdk.data_models.configurable_parameter import (
    ConfigurableBoolean,
    ConfigurableFloat,
    ConfigurableInteger,
    ConfigurableParameter,
    SelectableFloat,
    SelectableString,
)
from geti_sdk.data_models.enums.configuration_enums import (
    ConfigurableParameterType,
    ParameterDataType,
    ParameterInputType,
)
from geti_sdk.data_models.utils import (
    attr_value_serializer,
    deidentify,
    str_to_enum_converter,
)

PARAMETER_TYPES = Union[
    SelectableFloat,
    SelectableString,
    ConfigurableFloat,
    ConfigurableInteger,
    ConfigurableBoolean,
]


def _parameter_dicts_to_list(
    parameter_dicts: List[Union[Dict[str, Any], PARAMETER_TYPES]]
) -> List[PARAMETER_TYPES]:
    """
    Convert a list of dictionary representations of configurable parameters to a
    list of ConfigurableParameter objects.

    :param parameter_dicts: List of dictionaries, each entry representing a single
        configurable parameter
    :return: List of corresponding ConfigurableParameter objects
    """
    parameters: List[PARAMETER_TYPES] = []
    for parameter in parameter_dicts:
        if isinstance(parameter, get_args(PARAMETER_TYPES)):
            parameters.append(parameter)
            continue
        data_type = parameter.get("data_type", None)
        template_type = parameter.get("template_type", None)
        if data_type is None or template_type is None:
            raise ValueError(
                f"Unable to reconstruct ParameterGroup object from input: "
                f"{parameter_dicts}. No data or template type found for "
                f"parameter {parameter}"
            )
        data_type = ParameterDataType(data_type)
        template_type = ParameterInputType(template_type)
        parameter_type: Type[ConfigurableParameter]
        if (
            data_type == ParameterDataType.STRING
            and template_type == ParameterInputType.SELECTABLE
        ):
            parameter_type = SelectableString
        elif (
            data_type == ParameterDataType.INTEGER
            and template_type == ParameterInputType.INPUT
        ):
            parameter_type = ConfigurableInteger
        elif (
            data_type == ParameterDataType.BOOLEAN
            and template_type == ParameterInputType.INPUT
        ):
            parameter_type = ConfigurableBoolean
        elif (
            data_type == ParameterDataType.FLOAT
            and template_type == ParameterInputType.INPUT
        ):
            parameter_type = ConfigurableFloat
        elif (
            data_type == ParameterDataType.FLOAT
            and template_type == ParameterInputType.SELECTABLE
        ):
            parameter_type = SelectableFloat
        else:
            raise ValueError(
                f"Unable to reconstruct ParameterGroup object from input: "
                f"{parameter_dicts}. Invalid data or template type found "
                f"for parameter {parameter}"
            )
        parameters.append(parameter_type(**parameter))
    return parameters


@attr.define
class ParameterGroup:
    """
    Representation of a group of configurable parameters in GETi, as returned by the
    /configuration endpoints.

    :var header: Human readable name for the parameter group
    :var type: Type of the parameter group
    :var name: name by which the parameter group is identified in the system
    :var parameters: List of configurable parameters
    :var groups: List of parameter groups
    """

    _non_minimal_fields: ClassVar[List[str]] = ["description"]

    header: str
    type: str = attr.field(converter=str_to_enum_converter(ConfigurableParameterType))
    description: Optional[str] = None
    parameters: Optional[Sequence[PARAMETER_TYPES]] = None
    name: Optional[str] = None
    groups: Optional[List["ParameterGroup"]] = None

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the ParameterGroup
        """
        return attr.asdict(self, recurse=True, value_serializer=attr_value_serializer)

    @classmethod
    def from_dict(cls, input_dict: Dict[str, Any]) -> "ParameterGroup":
        """
        Create a ParameterGroup object from an input dictionary.

        :param input_dict: Dictionary to create the parameter group from
        :return: ParameterGroup instance corresponding to the data received in
            input_dict
        """
        input_copy = copy.deepcopy(input_dict)
        parameter_dicts: List[Union[Dict[str, Any], ConfigurableParameter]] = (
            input_copy.pop("parameters", [])
        )
        group_dicts: List[Union[Dict[str, Any], ParameterGroup]] = input_copy.pop(
            "groups", []
        )

        parameters = _parameter_dicts_to_list(parameter_dicts=parameter_dicts)
        groups: List[ParameterGroup] = []
        for group_dict in group_dicts:
            if isinstance(group_dict, ParameterGroup):
                groups.append(group_dict)
                continue
            groups.append(ParameterGroup.from_dict(group_dict))
        return cls(**input_copy, parameters=parameters, groups=groups)

    def deidentify(self) -> None:
        """
        Remove all identifier fields from the ParameterGroup
        """
        for parameter in self.parameters:
            deidentify(parameter)
        for group in self.groups:
            group.deidentify()

    def parameter_names(self, get_nested: bool = True) -> List[str]:
        """
        Return a list of names of all parameters in the ParameterGroup.

        :param get_nested: True to include parameter names in nested ParameterGroups
            belonging to this group as well. False to only consider this groups'
            parameters
        :return: List of parameter names
        """
        parameter_names = [parameter.name for parameter in self.parameters]
        if get_nested:
            for group in self.groups:
                parameter_names.extend(group.parameter_names(get_nested=True))
        return parameter_names

    def get_parameter_group_by_name(self, name: str) -> Optional["ParameterGroup"]:
        """
        Return the parameter group named `name`, if it belongs to this
        ParameterGroup or any of it's nested groups.

        If no group by `name` is found, this method returns None

        :param name: Name of the parameter group to look for
        :return: ParameterGroup named `name`, if any group by that name is found.
            None otherwise
        """
        all_groups = []
        for group in self.groups:
            all_groups.extend(group.groups)
        return next((group for group in all_groups if group.name == name), None)

    def get_parameter_by_name(
        self, name: str, group_name: Optional[str] = None
    ) -> Optional[PARAMETER_TYPES]:
        """
        Get the data for the configurable parameter named `name` from the
        ParameterGroup. This method returns None if no parameter by that name was found

        :param name: Name of the parameter to get
        :param group_name: Optional name of the parameter group to which the parameter
            belongs. If None is specified (the default), this method looks in all
            parameter groups belonging to this group.
        :return: ConfigurableParameter object representing the parameter named `name`
        """
        if name not in self.parameter_names(get_nested=True):
            return None

        if group_name is None:
            parameters = [
                parameter for parameter in self.parameters if parameter.name == name
            ]
            for group in self.groups:
                parameters.extend(
                    [
                        parameter
                        for parameter in group.parameters
                        if parameter.name == name
                    ]
                )
            if len(parameters) != 1:
                raise ValueError(
                    f"Found multiple parameters named {name}, please specify a "
                    f"group_name to identify the parameter unambiguously."
                )
            parameter = parameters[0]
        else:
            group = self.get_parameter_group_by_name(group_name)
            if group is not None:
                parameter = group.get_parameter_by_name(name)
            else:
                parameter = None
        return parameter

    def get_group_containing(self, parameter_name: str) -> Optional["ParameterGroup"]:
        """
        Return the parameter group that contains the parameter specified in
        `parameter_name`. If no group containing the parameter is found, this method
        returns None

        :param parameter_name: Name of the parameter for which to retrieve the group
        :return: ParameterGroup containing the parameter by name `parameter_name`
        """
        if not self.groups or self.get_parameter_by_name(parameter_name) is None:
            return None
        group_names = [group.name for group in self.groups]
        for group_name in group_names:
            if self.get_parameter_by_name(parameter_name, group_name) is not None:
                return self.get_parameter_group_by_name(group_name)

    @property
    def summary(self) -> str:
        """
        Return a string containing a very brief summary of the ParameterGroup.

        :return: string holding a very short summary of the ParameterGroup
        """
        summary_str = f"{self.header}"
        for parameter in self.parameters:
            summary_str += f"\n  {parameter.summary}"
        for group in self.groups:
            summary_str += f"\n  {group.summary}"
        summary_str += "\n"
        return summary_str



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
""" This module defines the test configuration """
import logging
import os
import shutil
import tempfile
from typing import Union

import pytest
from _pytest.main import Session

from geti_sdk import Geti
from geti_sdk.http_session import ServerCredentialConfig, ServerTokenConfig
from tests.helpers.project_helpers import remove_all_test_projects

from .helpers import SdkTestMode, get_sdk_fixtures, replace_unique_entries_in_cassettes
from .helpers.constants import (
    BASE_TEST_PATH,
    CASSETTE_BASE_PATH,
    CASSETTE_PATH_KEY,
    DUMMY_HOST,
    DUMMY_ORGANIZATION_ID,
    DUMMY_WORKSPACE_ID,
    RECORD_CASSETTE_KEY,
)

pytest_plugins = get_sdk_fixtures()

# -------------------------------------------------------
# ---------------- Environment variables ----------------
# -------------------------------------------------------

TEST_MODE = SdkTestMode[os.environ.get("TEST_MODE", "OFFLINE")]
# TEST_MODE specifies the mode in which tests are run. Only applies to the integration
# tests. Possible modes are: "OFFLINE", "ONLINE", "RECORD"

GETI_PLATFORM_VERSION = os.environ.get("GETI_PLATFORM_VERSION", "DEVELOP")
# GETI_PLATFORM_VERSION specifies the version of the Geti platform to run the tests
# against. This is only used in the nightly tests in "OFFLINE" and "RECORD" modes.
# Possible values: "DEVELOP", "LEGACY", "SAAS".

HOST = os.environ.get("GETI_HOST", "https://dummy_host").strip("/")
# HOST should hold the domain name or ip address of the Geti instance to run the tests
# against.

USERNAME = os.environ.get("GETI_USERNAME", "dummy_user")
# USERNAME should hold the username that is used for logging in to the Geti instance

PASSWORD = os.environ.get("GETI_PASSWORD", "dummy_password")
# PASSWORD should hold the password that is used for logging in to the Geti instance

TOKEN = os.environ.get("GETI_TOKEN", None)
# TOKEN should hold the Personal Access Token that can be used to access the server.
# When both a TOKEN and username + password are provided, the test suite will use the
# TOKEN to execute the tests

GETI_HTTP_PROXY = os.environ.get("GETI_HTTP_PROXY_URL", None)
GETI_HTTPS_PROXY = os.environ.get("GETI_HTTPS_PROXY_URL", None)
# GETI_HTTP_PROXY and GETI_HTTPS_PROXY are urls to the proxy servers that should be
# used to connect to the Geti instance.
# NOTE: PROXIES can only be used in ONLINE mode, they cannot be used in RECORD mode
# (will raise an error) and have no effect in OFFLINE mode.

CLEAR_EXISTING_TEST_PROJECTS = os.environ.get(
    "CLEAR_EXISTING_TEST_PROJECTS", "0"
).lower() in ["true", "1"]
# CLEAR_EXISTING_TEST_PROJECTS is a boolean that determines whether existing test
# projects are deleted before a test run

NIGHTLY_TEST_LEARNING_PARAMETER_SETTINGS = os.environ.get(
    "LEARNING_PARAMETER_SETTINGS", "default"
)
# NIGHTLY_TEST_LEARNING_PARAMETER_SETTINGS determines how the learning parameters are
# set for the nightly tests. Possible values are:
#   - "default"     : The default settings
#   - "minimal"     : Single epoch and batch_size of 1
#   - "reduced_mem" : Default epochs, but batch_size of 1

# ------------------------------------------
# ---------------- Fixtures ----------------
# ------------------------------------------


@pytest.fixture(scope="session")
def fxt_server_config() -> Union[ServerTokenConfig, ServerCredentialConfig]:
    """
    This fixture holds the login configuration to access the Geti server
    """
    # Configure proxies for OFFLINE, ONLINE and RECORD mode
    if TEST_MODE == SdkTestMode.OFFLINE:
        proxies = {"https": "", "http": ""}
    elif TEST_MODE == SdkTestMode.ONLINE and (
        GETI_HTTP_PROXY is not None or GETI_HTTPS_PROXY is not None
    ):
        proxies = {"https": GETI_HTTPS_PROXY, "http": GETI_HTTP_PROXY}
    else:
        # In RECORD mode or when both proxies are None, the `proxies` argument for the
        # ServerConfig should be set to None
        proxies = None
        if TEST_MODE == SdkTestMode.RECORD and (
            GETI_HTTPS_PROXY is not None or GETI_HTTP_PROXY is not None
        ):
            raise ValueError(
                "Unable to use proxy servers in RECORD mode! Please clear the "
                "GETI_HTTPS_PROXY and GETI_HTTP_PROXY environment variables before "
                "running the test suite in RECORD mode."
            )

    # Use token if available
    if TOKEN is None:
        test_config = ServerCredentialConfig(
            host=HOST, username=USERNAME, password=PASSWORD, proxies=proxies
        )
    else:
        test_config = ServerTokenConfig(host=HOST, token=TOKEN, proxies=proxies)
    yield test_config


@pytest.fixture(scope="session")
def fxt_base_test_path() -> str:
    """
    This fixture returns the absolute path to the `tests` folder
    """
    yield BASE_TEST_PATH


@pytest.fixture(scope="session")
def fxt_test_mode() -> SdkTestMode:
    """
    This fixture returns the SdkTestMode with which the tests are run
    :return:
    """
    yield TEST_MODE


@pytest.fixture(scope="session")
def fxt_learning_parameter_settings() -> str:
    """
    This fixture returns the settings to use in the nightly test learning parameters.
    Can be either:
      - 'minimal' (single epoch, batch_size = 1),
      - 'reduced_mem' (normal epochs but reduced batch size for memory hungry algos)
      - 'default' (default settings)

    :return:
    """
    yield NIGHTLY_TEST_LEARNING_PARAMETER_SETTINGS


@pytest.fixture(scope="session")
def fxt_github_actions_environment() -> bool:
    """
    Return True if the tests are running in a GitHub actions environment, False
    otherwise
    """
    yield os.environ.get("GITHUB_ACTIONS", False)


# ----------------------------------------------
# ---------------- Pytest hooks ----------------
# ----------------------------------------------


def _get_geti_instance() -> Geti:
    """
    This function returns a Geti instance with the correct authentication parameters
    """
    # Handle authentication via token or credentials
    if TOKEN is None:
        auth_params = {"username": USERNAME, "password": PASSWORD}
    else:
        auth_params = {"token": TOKEN}
    # Handle proxies
    if GETI_HTTP_PROXY is not None or GETI_HTTPS_PROXY is not None:
        proxies = {"http": GETI_HTTP_PROXY, "https": GETI_HTTPS_PROXY}
    else:
        proxies = None
    return Geti(host=HOST, **auth_params, proxies=proxies, verify_certificate=False)


def pytest_sessionstart(session: Session) -> None:
    """
    This function is called before a pytest test run begins.

    If the tests are run in record mode, this hook sets up a temporary directory to
    record the new cassettes to.

    :param session: Pytest session instance that has just been created
    """
    versioned_cassette_path = os.path.join(CASSETTE_BASE_PATH, GETI_PLATFORM_VERSION)
    if not os.path.exists(versioned_cassette_path):
        os.makedirs(versioned_cassette_path)
    os.environ.update({CASSETTE_PATH_KEY: versioned_cassette_path})
    if CLEAR_EXISTING_TEST_PROJECTS and TEST_MODE != SdkTestMode.OFFLINE:
        # Remove existing test projects
        remove_all_test_projects(geti=_get_geti_instance())
    if TEST_MODE == SdkTestMode.RECORD:
        record_cassette_path = tempfile.mkdtemp()
        logging.info(f"Cassettes will be recorded to `{record_cassette_path}`.")
        os.environ.update({RECORD_CASSETTE_KEY: record_cassette_path})


def pytest_sessionfinish(session: Session, exitstatus: int) -> None:
    """
    This function is called after a pytest test run finishes.

    If the tests are run in record mode, this hook handles saving and cleaning the
    recorded cassettes

    :param session: Pytest session that has just finished
    :param exitstatus: Exitstatus with which the tests completed
    """
    if TEST_MODE == SdkTestMode.RECORD:
        record_cassette_path = os.environ.pop(RECORD_CASSETTE_KEY)
        versioned_cassette_path = os.environ.pop(CASSETTE_PATH_KEY)
        if exitstatus == 0:
            logging.info("Recording successful! Wrapping up....")
            # Scrub hostname, organization_id and workspace_id from cassettes
            geti = _get_geti_instance()
            organization_id_pair = (
                geti.session._organization_id,
                DUMMY_ORGANIZATION_ID,
            )
            workspace_id_pair = (geti.workspace_id, DUMMY_WORKSPACE_ID)
            host_name_pair = (HOST.replace("https://", "").strip("/"), DUMMY_HOST)

            # replace_host_name_in_cassettes(HOST, cassette_dir=record_cassette_path)
            replace_unique_entries_in_cassettes(
                (organization_id_pair, workspace_id_pair, host_name_pair),
                cassette_dir=record_cassette_path,
            )
            logging.info(
                " Hostname, organization and workspace ids were scrubbed from all cassette files successfully."
            )
            # Copy recorded cassettes to fixtures/cassettes
            logging.info(
                f"Copying newly recorded cassettes from `{record_cassette_path}` to "
                f"`{versioned_cassette_path}`."
            )
            for root, dirs, files in os.walk(record_cassette_path):
                for file in files:
                    shutil.move(
                        os.path.join(root, file),
                        os.path.join(versioned_cassette_path, file),
                    )

        else:
            # Clean up any cassettes already recorded
            if os.path.exists(record_cassette_path):
                shutil.rmtree(record_cassette_path)



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
from typing import Any, Dict, List, Union

import attr

from geti_sdk.data_models.configurable_parameter_group import ParameterGroup
from geti_sdk.data_models.configuration import (
    ConfigurableParameters,
    FullConfiguration,
    GlobalConfiguration,
    TaskConfiguration,
)
from geti_sdk.data_models.configuration_identifiers import (
    ComponentEntityIdentifier,
    EntityIdentifier,
    HyperParameterGroupIdentifier,
)
from geti_sdk.data_models.enums.configuration_enums import ConfigurationEntityType
from geti_sdk.data_models.utils import remove_null_fields


class ConfigurationRESTConverter:
    """
    Class that handles conversion of Intel® Geti™ REST output for configurable parameter
    entities to objects and vice versa.
    """

    @staticmethod
    def entity_identifier_from_dict(input_dict: Dict[str, Any]) -> EntityIdentifier:
        """
        Create an EntityIdentifier object from an input dictionary.

        :param input_dict: Dictionary representing an EntityIdentifier in Intel® Geti™
        :return: EntityIdentifier object corresponding to the data in `input_dict`
        """
        identifier_type = input_dict.get("type", None)
        if isinstance(identifier_type, str):
            identifier_type = ConfigurationEntityType(identifier_type)
        if identifier_type == ConfigurationEntityType.HYPER_PARAMETER_GROUP:
            identifier_class = HyperParameterGroupIdentifier
        elif identifier_type == ConfigurationEntityType.COMPONENT_PARAMETERS:
            identifier_class = ComponentEntityIdentifier
        else:
            raise ValueError(
                f"Invalid entity identifier type found: Entity identifier of type "
                f"{identifier_type} is not supported."
            )
        return identifier_class(**input_dict)

    @staticmethod
    def from_dict(input_dict: Dict[str, Any]) -> ConfigurableParameters:
        """
        Create a ConfigurableParameters object holding the configurable parameters
        for an entity in the Intel® Geti™ platform, from a dictionary returned by the
        /configuration REST endpoints.

        :param input_dict: Dictionary containing the configurable parameters
        :return: ConfigurableParameters instance holding the parameter data
        """
        input_copy = copy.deepcopy(input_dict)
        entity_identifier = input_copy.pop("entity_identifier")

        if not isinstance(entity_identifier, EntityIdentifier):
            entity_identifier = ConfigurationRESTConverter.entity_identifier_from_dict(
                input_dict=entity_identifier
            )

        input_copy["entity_identifier"] = entity_identifier
        return ConfigurableParameters.from_dict(input_dict=input_copy)

    @staticmethod
    def _rest_components_to_objects(
        input_list: List[Dict[str, Any]]
    ) -> List[ConfigurableParameters]:
        """
        Create a list of configurable parameters from a list of dictionaries received
        by the Intel® Geti™ /configuration endpoints.

        :param input_list: List of dictionaries to convert
        :return: List of ConfigurableParameters instances
        """
        component_objects: List[ConfigurableParameters] = []
        for component in input_list:
            if not isinstance(component, ConfigurableParameters):
                component_objects.append(
                    ConfigurationRESTConverter.from_dict(component)
                )
            else:
                component_objects.append(component)
        return component_objects

    @staticmethod
    def _remove_non_minimal_fields(configurable_parameters: ParameterGroup):
        """
        For all parameters in the parameter group, set the fields that are not part
        of the minimal representation of the parameter to 'None'.

        NOTE: This method modifies the input in place

        :param configurable_parameters: ConfigurableParameters instance to convert to
            minimal representation
        :return:
        """
        # Set non minimal fields of ConfigurableParameters object to None
        for field in attr.fields(type(configurable_parameters)):
            if field.name in configurable_parameters._non_minimal_fields:
                setattr(configurable_parameters, field.name, None)
        # Set non minimal fields of individual parameters to None
        for parameter in configurable_parameters.parameters:
            for field in attr.fields(type(parameter)):
                if field.name in parameter._non_minimal_fields:
                    setattr(parameter, field.name, None)
        # Set groups to None if no groups, else remove redundant fields from all groups
        if not configurable_parameters.groups:
            configurable_parameters.groups = None
        else:
            for group in configurable_parameters.groups:
                ConfigurationRESTConverter._remove_non_minimal_fields(group)

    @staticmethod
    def task_configuration_from_dict(input_dict: Dict[str, Any]) -> TaskConfiguration:
        """
        Create a TaskConfiguration object holding all configurable parameters for a
        task in an Intel® Geti™ project, from a dictionary returned by the
        /configuration REST endpoints.

        :param input_dict: Dictionary containing the configurable parameters for the
            task
        :return: TaskConfiguration instance holding the parameter data
        """
        input_copy = copy.deepcopy(input_dict)
        components = input_copy.pop("components")
        component_objects = ConfigurationRESTConverter._rest_components_to_objects(
            components
        )
        input_copy.update({"components": component_objects})
        return TaskConfiguration(**input_copy)

    @staticmethod
    def configuration_to_minimal_dict(
        configuration: Union[TaskConfiguration, GlobalConfiguration, FullConfiguration],
        deidentify: bool = True,
    ) -> Dict[str, Any]:
        """
        Convert a TaskConfiguration, GlobalConfiguration or FullConfiguration into a
        dictionary, removing fields that are None or are only relevant to the
        Intel® Geti™ UI.

        :param configuration: TaskConfiguration or GlobalConfiguration to convert
        :param deidentify: True to remove all unique database identifiers, False to
            preserve identifiers. Defaults to True
        :return: Dictionary representation of the configuration
        """
        input_copy = copy.deepcopy(configuration)
        if deidentify:
            input_copy.deidentify()
        if not isinstance(configuration, FullConfiguration):
            for config in input_copy.components:
                ConfigurationRESTConverter._remove_non_minimal_fields(config)
        else:
            for config in input_copy.global_.components:
                ConfigurationRESTConverter._remove_non_minimal_fields(config)
            for task_config in input_copy.task_chain:
                for config in task_config.components:
                    ConfigurationRESTConverter._remove_non_minimal_fields(config)
        result = input_copy.to_dict()
        remove_null_fields(result)
        return result

    @staticmethod
    def global_configuration_from_rest(
        input_: Union[List[Dict[str, Any]], Dict[str, Any]]
    ) -> GlobalConfiguration:
        """
        Create a GlobalConfiguration object holding the configurable parameters
        for all project-wide components in the Intel® Geti™ project, from input from the
        /configuration/global REST endpoint.

        :param input_: REST response holding the serialized configurable parameters
        :return:
        """
        input_copy = copy.deepcopy(input_)
        if isinstance(input_copy, list):
            input_list = input_copy
        else:
            input_list = input_copy.pop("items")
        component_objects = ConfigurationRESTConverter._rest_components_to_objects(
            input_list
        )
        if isinstance(input_copy, list):
            return GlobalConfiguration(components=component_objects)
        else:
            input_copy.update({"components": component_objects})
            return GlobalConfiguration(**input_copy)

    @staticmethod
    def full_configuration_from_rest(input_dict: Dict[str, Any]) -> FullConfiguration:
        """
        Convert a dictionary holding the full configuration for an Intel® Geti™
        project, as returned by the /configuration endpoint, to an object
        representation.

        :param input_dict: Dictionary representing the full project configuration
        :return: FullConfiguration instance holding the global and task chain
            configuration
        """
        global_dict = input_dict.pop("global")
        task_chain_list = input_dict.pop("task_chain")
        global_config = ConfigurationRESTConverter.global_configuration_from_rest(
            global_dict
        )
        task_chain_config = [
            ConfigurationRESTConverter.task_configuration_from_dict(task_config)
            for task_config in task_chain_list
        ]
        return FullConfiguration(global_=global_config, task_chain=task_chain_config)

    @staticmethod
    def configurable_parameter_list_to_rest(
        configurable_parameter_list: List[ConfigurableParameters],
    ) -> Dict[str, List[Dict[str, Any]]]:
        """
        Convert a list of model hyper parameters to a dictionary that can be sent to
        the /configuration POST endpoints.

        :param configurable_parameter_list: List of ConfigurableParameter instances
        :return: Dictionary containing:
            - 'components': list of dictionaries representing configurable parameters,
                            that are conforming to the /configuration REST endpoints
        """
        rest_parameters: List[Dict[str, Any]] = []
        for parameter_set in configurable_parameter_list:
            parameter_copy = copy.deepcopy(parameter_set)
            ConfigurationRESTConverter._remove_non_minimal_fields(parameter_copy)
            parameter_dict = parameter_copy.to_dict()
            remove_null_fields(parameter_dict)
            rest_parameters.append(parameter_dict)
        return {"components": rest_parameters}



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `data_models` package contains the SDK representation for all entities in GETi, such
as :py:class:`~geti_sdk.data_models.annotation_scene.AnnotationScene`,
:py:class:`~geti_sdk.data_models.media.Image`,
:py:class:`~geti_sdk.data_models.project.Project` and
:py:class:`~geti_sdk.data_models.model.Model` and many more.

When interacting with the GETi cluster through the
:py:class:`geti_sdk.sc_rest_client.Geti` or the
:py:mod:`~geti_sdk.rest_clients`, all entities retrieved from the cluster will be
deserialized into the data models defined in this package.

Module contents
---------------

Algorithm-related entities
++++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.algorithms
   :members:
   :undoc-members:

Project-related entities
++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.task
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.project
   :members:
   :undoc-members:

Annotation-related entities
+++++++++++++++++++++++++++
.. automodule:: geti_sdk.data_models.label
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.shapes
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.annotations
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.annotation_scene
   :members:
   :undoc-members:

Configuration-related entities
++++++++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.configurable_parameter
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.configurable_parameter_group
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.configuration_identifiers
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.configuration
   :members:
   :undoc-members:

Model-related entities
++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.model
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.model_group
   :members:
   :undoc-members:

Media-related entities
++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.media
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.media_identifiers
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.task_annotation_state
   :members:
   :undoc-members:

Prediction-related entities
+++++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.predictions
   :members:
   :undoc-members:

Status- and job-related entities
++++++++++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.status
   :members:
   :undoc-members:

.. automodule:: geti_sdk.data_models.job
   :members:
   :undoc-members:

Deployment-related entities
+++++++++++++++++++++++++++

.. automodule:: geti_sdk.data_model.code_deployment_info
   :members:
   :undoc-members:

Utility functions
+++++++++++++++++

.. automodule:: geti_sdk.data_models.utils
   :members:
   :undoc-members:

Custom container classes
++++++++++++++++++++++++

.. automodule:: geti_sdk.data_models.containers
   :members:
   :undoc-members:

Enumerations
++++++++++++

.. automodule:: geti_sdk.data_models.enums
   :members:
   :undoc-members:

"""

from .algorithms import Algorithm
from .annotation_scene import AnnotationScene
from .annotations import Annotation
from .code_deployment_info import CodeDeploymentInformation
from .configuration import (
    ConfigurableParameters,
    FullConfiguration,
    GlobalConfiguration,
    TaskConfiguration,
)
from .enums import AnnotationKind, MediaType, TaskType
from .job import Job
from .label import Label, ScoredLabel
from .media import Image, MediaItem, Video, VideoFrame
from .model import Model, OptimizedModel
from .model_group import ModelGroup, ModelSummary
from .performance import Performance
from .predictions import Prediction
from .project import Dataset, Pipeline, Project
from .status import ProjectStatus
from .task import Task
from .test_result import Score, TestResult
from .user import User

__all__ = [
    "TaskType",
    "AnnotationKind",
    "Project",
    "Label",
    "Task",
    "Pipeline",
    "Image",
    "Video",
    "MediaItem",
    "MediaType",
    "Annotation",
    "AnnotationScene",
    "Algorithm",
    "ScoredLabel",
    "VideoFrame",
    "Prediction",
    "Performance",
    "TaskConfiguration",
    "GlobalConfiguration",
    "ConfigurableParameters",
    "FullConfiguration",
    "Model",
    "ModelGroup",
    "OptimizedModel",
    "ModelSummary",
    "ProjectStatus",
    "Job",
    "CodeDeploymentInformation",
    "Dataset",
    "TestResult",
    "Score",
    "User",
]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import json
import logging
import os
import time
from typing import Any, Dict, List, Optional, Tuple, Union

from geti_sdk.data_models import Project, Task, TaskType
from geti_sdk.data_models.utils import remove_null_fields
from geti_sdk.http_session import GetiRequestException, GetiSession
from geti_sdk.rest_clients.dataset_client import DatasetClient
from geti_sdk.rest_converters import ProjectRESTConverter
from geti_sdk.utils.label_helpers import generate_unique_label_color
from geti_sdk.utils.project_helpers import get_task_types_by_project_type

from .task_templates import (
    ANOMALY_CLASSIFICATION_TASK,
    ANOMALY_DETECTION_TASK,
    ANOMALY_SEGMENTATION_TASK,
    BASE_TEMPLATE,
    CLASSIFICATION_TASK,
    CROP_TASK,
    DETECTION_TASK,
    INSTANCE_SEGMENTATION_TASK,
    ROTATED_DETECTION_TASK,
    SEGMENTATION_TASK,
)

TASK_TYPE_MAPPING = {
    TaskType.CROP: CROP_TASK,
    TaskType.DETECTION: DETECTION_TASK,
    TaskType.SEGMENTATION: SEGMENTATION_TASK,
    TaskType.CLASSIFICATION: CLASSIFICATION_TASK,
    TaskType.ANOMALY_CLASSIFICATION: ANOMALY_CLASSIFICATION_TASK,
    TaskType.ANOMALY_DETECTION: ANOMALY_DETECTION_TASK,
    TaskType.ANOMALY_SEGMENTATION: ANOMALY_SEGMENTATION_TASK,
    TaskType.INSTANCE_SEGMENTATION: INSTANCE_SEGMENTATION_TASK,
    TaskType.ROTATED_DETECTION: ROTATED_DETECTION_TASK,
}


class ProjectClient:
    """
    Class to manipulate projects on the Intel® Geti™ server, within a certain workspace.
    """

    def __init__(self, session: GetiSession, workspace_id: str):
        self.session = session
        self.workspace_id = workspace_id
        self.base_url = f"workspaces/{workspace_id}/"

    def get_all_projects(
        self, request_page_size: int = 50, get_project_details: bool = True
    ) -> List[Project]:
        """
        Return a list of projects found on the Intel® Geti™ server

        :param request_page_size: Max number of projects to fetch in a single HTTP
            request. Higher values may reduce the response time of this method when
            there are many projects, but increase the chance of timeout.
        :param get_project_details: True to get all details of the projects on the
            Intel® Geti™, False to fetch only a summary of each project. Set this to
            False if minimizing latency is a concern. Defaults to True
        :return: List of Project objects, containing the project information for each
            project on the Intel® Geti™ server
        """
        project_key = "projects"
        num_total_projects_key = "project_counts"

        # The 'projects' endpoint uses pagination: multiple HTTP may be necessary to
        # fetch the full list of projects
        project_rest_list: List[Dict] = []
        while response := self.session.get_rest_response(
            url=f"{self.base_url}projects?limit={request_page_size}&skip={len(project_rest_list)}",
            method="GET",
        ):
            project_rest_list.extend(response[project_key])
            if len(project_rest_list) >= response[num_total_projects_key]:
                break

        project_list = [
            ProjectRESTConverter.from_dict(project_input=project)
            for project in project_rest_list
        ]
        if get_project_details:
            project_detail_list: List[Project] = []
            for project in project_list:
                try:
                    project_detail_list.append(self._get_project_detail(project))
                except GetiRequestException as e:
                    if e.status_code == 403:
                        logging.info(
                            f"Unable to access project `{project.name}` details: Unauthorized."
                        )
                        project_detail_list.append(project)
            return project_detail_list
        else:
            return project_list

    def get_project_by_name(
        self, project_name: str, project_id: Optional[str] = None
    ) -> Optional[Project]:
        """
        Get a project from the Intel® Geti™ server by project_name.

        :param project_name: Name of the project to get
        :param project_id: Optional ID of the project to get. Only used if more than
            one project named `project_name` exists in the workspace.
        :raises: ValueError in case multiple projects with the specified name exist on
            the server, and no `project_id` is provided in order to allow unique
            identification of the project.
        :return: Project object containing the data of the project, if the project is
            found on the server. Returns None if the project doesn't exist
        """
        project_list = self.get_all_projects(get_project_details=False)
        matches = [project for project in project_list if project.name == project_name]
        if len(matches) == 1:
            return self._get_project_detail(matches[0])
        elif len(matches) > 1:
            if project_id is None:
                detailed_matches = [
                    self._get_project_detail(match) for match in matches
                ]
                projects_info = [
                    (
                        f"Name: {p.name}, Type: {p.project_type}, ID: {p.id}, "
                        f"creation_date: {p.creation_time}"
                    )
                    for p in detailed_matches
                ]
                raise ValueError(
                    f"A total of {len(matches)} projects named `{project_name}` were "
                    f"found in the workspace. Unable to uniquely identify the "
                    f"desired project. Please provide a `project_id` to ensure the "
                    f"proper project is returned. The following projects were found:"
                    f"{projects_info}"
                )
            else:
                matched_project = next(
                    (project for project in matches if project.id == project_id), None
                )
                if matched_project is None:
                    logging.info(
                        f"Projects with name `{project_name}` were found, but none of "
                        f"the project ID's `{[p.id for p in matches]}` matches the "
                        f"requested id `{project_id}`."
                    )
                return matched_project
        else:
            return None

    def get_or_create_project(
        self,
        project_name: str,
        project_type: str,
        labels: List[Union[List[str], List[Dict[str, Any]]]],
    ) -> Project:
        """
        Create a new project with name `project_name` on the cluster, or retrieve
        the data for an existing project with `project_name` if it exists.

        :param project_name: Name of the project
        :param project_type: Type of the project
        :param labels: Nested list of labels
        :return:
        """
        project = self.get_project_by_name(project_name)
        if project is not None:
            logging.info(
                f"Project with name {project_name} exists, continuing with "
                f"exiting project. No new project has been created."
            )
        else:
            project = self.create_project(
                project_name=project_name, project_type=project_type, labels=labels
            )
        return project

    def create_project(
        self,
        project_name: str,
        project_type: str,
        labels: List[Union[List[str], List[Dict[str, Any]]]],
    ) -> Project:
        """
        Create a new project with name `project_name` on the cluster, containing
        tasks according to the `project_type` specified. Labels for each task are
        specified in the `labels` parameter, which should be a nested list (each entry
        in the outermost list corresponds to the labels for one of the tasks in the
        project pipeline)

        :param project_name: Name of the project
        :param project_type: Type of the project
        :param labels: Nested list of labels
        :raises ValueError: If a project with name `project_name` already exists in
            the workspace
        :return: Project object, as created on the cluster
        """
        project_template = self._create_project_template(
            project_name=project_name, project_type=project_type, labels=labels
        )
        project = self.session.get_rest_response(
            url=f"{self.base_url}projects", method="POST", data=project_template
        )
        logging.info("Project created successfully.")
        project = ProjectRESTConverter.from_dict(project)
        self._await_project_ready(project=project)
        return project

    def download_project_info(self, project_name: str, path_to_folder: str) -> None:
        """
        Get the project data that can be used for project creation for a project on
        the Intel® Geti™ server, named `project_name`. From the returned data, the
        method `ProjectClient.get_or_create_project` can create a project on the
        Intel® Geti™ server. The data is retrieved from the cluster and saved in the
        target folder `path_to_folder`.

        :param project_name: Name of the project to retrieve the data for
        :param path_to_folder: Target folder to save the project data to.
            Data will be saved as a .json file named "project.json"
        :raises ValueError: If the project with `project_name` is not found on the
            cluster
        """
        project = self.get_project_by_name(project_name)
        if project is None:
            raise ValueError(
                f"Project with name {project_name} was not found on the server."
            )
        project_data = ProjectRESTConverter.to_dict(project)
        os.makedirs(path_to_folder, exist_ok=True, mode=0o770)
        project_config_path = os.path.join(path_to_folder, "project.json")
        with open(project_config_path, "w") as file:
            json.dump(project_data, file, indent=4)
        logging.info(
            f"Project parameters for project '{project_name}' were saved to file "
            f"{project_config_path}."
        )

    @staticmethod
    def _add_task(
        project_template: dict,
        task_type: TaskType,
        labels: Union[List[str], List[Dict[str, Any]]],
    ) -> Tuple[dict, dict]:
        """
        Add a task to the pipeline in a project template in dictionary form.

        :param project_template: Dictionary representing the project creation data
        :param task_type: Type of the task to be added
        :param labels: Labels to be used for the task
        :return:
        """
        new_template = copy.deepcopy(project_template)
        tasks = new_template["pipeline"]["tasks"]
        try:
            task_template = copy.deepcopy(TASK_TYPE_MAPPING[task_type])
        except KeyError as error:
            raise ValueError(
                f"Task of type {task_type} is currently not supported."
            ) from error

        # Make sure the task title is unique in the pipeline
        task_titles_in_template = [task["title"] for task in tasks]
        unique_task_title = ProjectClient._ensure_unique_task_name(
            task_template["title"], task_titles_in_template
        )
        task_template["title"] = unique_task_title

        if not task_type.is_anomaly:
            label_group_name = f"{unique_task_title.lower()} label group"

            for label in labels:
                if isinstance(label, str):
                    label_info = {"name": label, "group": label_group_name}
                else:
                    label_info = label
                task_template["labels"].append(label_info)
        tasks.append(task_template)
        return new_template, task_template

    @staticmethod
    def _add_crop_task(project_template: dict) -> Tuple[dict, dict]:
        """
        Add a `crop` task to the pipeline in the project_template.

        :param project_template:
        :return: Tuple containing:
            - A dictionary representing the new project_template, with the crop task
                added to it
            - A dictionary representing the crop task that was added to the template
        """
        new_template = copy.deepcopy(project_template)
        tasks = new_template["pipeline"]["tasks"]
        crop_task = copy.deepcopy(CROP_TASK)
        tasks.append(crop_task)
        return new_template, crop_task

    @staticmethod
    def _add_connection(project_template: dict, to_task: str, from_task: str) -> dict:
        """
        Add a connection between `from_task` and `to_task` in the project_template
        dictionary.

        :param project_template:
        :param to_task:
        :param from_task:
        :return:
        """
        new_template = copy.deepcopy(project_template)
        connections = new_template["pipeline"]["connections"]
        connections.append({"from": from_task, "to": to_task})
        return new_template

    def create_project_from_folder(
        self, path_to_folder: str, project_name: Optional[str] = None
    ) -> Project:
        """
        Look for a `project.json` file in the folder at `path_to_folder`, and
        create a project using the parameters provided in this file.

        :param path_to_folder: Folder holding the project data
        :param project_name: Optional name of the project. If not specified, the
            project name found in the project configuration in the upload folder
            will be used.
        :return: Project as created on the cluster
        """
        path_to_project = os.path.join(path_to_folder, "project.json")
        if not os.path.isfile(path_to_project):
            raise ValueError(
                f"Unable to find project configuration file at {path_to_project}. "
                f"Please provide a valid path to the folder holding the project data."
            )
        with open(path_to_project, "r") as file:
            project_data = json.load(file)
        project = ProjectRESTConverter.from_dict(project_data)
        if project_name is not None:
            project.name = project_name
        logging.info(
            f"Creating project '{project.name}' from parameters in "
            f"configuration file at {path_to_project}."
        )
        project.prepare_for_post()
        datasets = project.datasets
        created_project = self.get_or_create_project(**project.get_parameters())
        if len(datasets) > 1:
            # Create the additional datasets if needed
            dataset_client = DatasetClient(
                session=self.session,
                workspace_id=self.workspace_id,
                project=created_project,
            )
            for dataset in datasets:
                if dataset.name not in [ds.name for ds in created_project.datasets]:
                    dataset_client.create_dataset(name=dataset.name)
        return created_project

    @staticmethod
    def is_project_dir(path_to_folder: str) -> bool:
        """
        Check if the folder specified in `path_to_folder` is a directory
        containing valid Intel® Geti™ project data that can be used to upload to an
        Intel® Geti™ server.

        :param path_to_folder: Directory to check
        :return: True if the directory holds project data, False otherwise
        """
        if not os.path.isdir(path_to_folder):
            return False
        path_to_project = os.path.join(path_to_folder, "project.json")
        if not os.path.isfile(path_to_project):
            return False
        try:
            with open(path_to_project, "r") as file:
                project_data = json.load(file)
        except json.decoder.JSONDecodeError:
            return False
        try:
            ProjectRESTConverter.from_dict(project_data)
        except (ValueError, TypeError, KeyError):
            return False
        return True

    def list_projects(self) -> List[Project]:
        """
        Print an overview of all projects that currently exists on the
        cluster, in the workspace managed by the ProjectClient

        NOTE: While this method also returns a list of all the projects, it is
        primarily meant to be used in an interactive environment, such as a
        Jupyter Notebook.

        :return: List of all Projects on the cluster. The returned list is the same as
            the list returned by the `get_all_projects` method
        """
        projects = self.get_all_projects()
        print(f"{len(projects)} projects were found on the platform:\n")
        for project in projects:
            print(" " + project.summary + "\n")
        return projects

    def _create_project_template(
        self,
        project_name: str,
        project_type: str,
        labels: List[Union[List[str], List[Dict[str, Any]]]],
    ) -> Dict[str, Any]:
        """
        Create a template dictionary with data for project creation that is ready to
        be sent to  the cluster.

        :param project_name: Name of the project
        :param project_type: Type of the project
        :param labels: Nested list of labels
        :return: Dictionary containing the data to create a project named
            `project_name`, of type `project_type` and with labels `labels`
        """
        project_template = copy.deepcopy(BASE_TEMPLATE)
        previous_task_name = "Dataset"
        previous_task_type = TaskType.DATASET
        task_names_in_template: List[str] = [previous_task_name]
        is_first_task = True
        for task_type, task_labels in zip(
            get_task_types_by_project_type(project_type), labels
        ):
            if not is_first_task and not previous_task_type.is_global:
                # Add crop task and connections, only for tasks that are not
                # first in the pipeline and are not preceded by a global task
                project_template, crop_task = self._add_crop_task(project_template)
                unique_task_name = self._ensure_unique_task_name(
                    crop_task["title"], task_names_in_template
                )
                crop_task["title"] = unique_task_name

                project_template = self._add_connection(
                    project_template,
                    to_task=unique_task_name,
                    from_task=previous_task_name,
                )
                previous_task_name = unique_task_name
            project_template, added_task = self._add_task(
                project_template, task_type=task_type, labels=task_labels
            )
            task_name = added_task["title"]

            project_template = self._add_connection(
                project_template, to_task=task_name, from_task=previous_task_name
            )
            previous_task_name = task_name
            previous_task_type = task_type
            is_first_task = False

        project_template["name"] = project_name
        return project_template

    @staticmethod
    def _ensure_unique_task_name(
        task_name: str, task_names_in_template: List[str]
    ) -> str:
        """
        Check that the `task_name` passed is not already in the list of
        `task_names_in_template`. If the task_name is already in the list, this method
        will generate a new task_name that is unique

        NOTE: This method updates the list of task_names_in_template with the unique
        task_name that is ensured in this method. This update is done in-place

        :param task_name: Task name to check
        :param task_names_in_template: List of task names that are already taken in the
            project template
        :return: Name of the task that is guaranteed to be unique in the template
        """
        ii = 2
        if task_name in task_names_in_template:
            new_task_name = f"{task_name} {ii}"
            while new_task_name in task_names_in_template:
                ii += 1
                new_task_name = f"{task_name} {ii}"
            task_name = new_task_name
        task_names_in_template.append(task_name)
        return task_name

    def delete_project(
        self, project: Union[str, Project], requires_confirmation: bool = True
    ) -> None:
        """
        Delete a project. The `project` to delete can either by a Project object or a
        string containing the name of the project to delete.

        By default, this method will ask for user confirmation before deleting the
        project. This can be overridden by passing `requires_confirmation = False`.

        :param project: Project to delete, either a string containing the project
            name or a Project instance
        :param requires_confirmation: True to ask for user confirmation before
            deleting the project, False to delete without confirmation. Defaults to
            True
        """
        if isinstance(project, str):
            project = self.get_project_by_name(project_name=project)
        if not isinstance(project, Project):
            raise TypeError(f"{type(project)} is not a valid project type.")

        if requires_confirmation:
            media_response = self.session.get_rest_response(
                url=f"{self.base_url}projects/{project.id}/datasets/"
                f"{project.datasets[0].id}/media",
                method="GET",
            )

            media_count = media_response.get("media_count", {"images": 0, "videos": 0})
            user_confirmation = input(
                f"CAUTION: You are about to delete project '{project.name}', "
                f"containing {media_count['images']} images and {media_count['videos']}"
                f" videos, from the platform. Are you sure you want to continue? Type "
                f"Y or YES to continue, any other key to cancel."
            )
            if not (
                user_confirmation.lower() == "yes" or user_confirmation.lower() == "y"
            ):
                logging.info("Aborting project deletion.")
                return
        try:
            self.session.get_rest_response(
                url=f"{self.base_url}projects/{project.id}", method="DELETE"
            )
        except GetiRequestException as error:
            if error.status_code == 409:
                raise ValueError(
                    f"Project {project.name} is locked for deletion/modification. "
                    f"Please wait until all jobs related to this project are finished "
                    f"or cancel them to allow deletion/modification."
                )
            else:
                raise error
        logging.info(f"Project '{project.name}' deleted successfully.")

    def add_labels(
        self,
        labels: Union[List[str], List[Dict[str, Any]]],
        project: Project,
        task: Optional[Task] = None,
        revisit_affected_annotations: bool = False,
    ) -> Project:
        """
        Add the `labels` to the project labels. For a project with multiple tasks,
        the `task` parameter can be used to specify the task for which the labels
        should be added.

        :param labels: List of labels to add. Can either be a list of strings
            representing label names, or a list of dictionaries representing label
            properties
        :param project: Project to which the labels should be added
        :param task: Optional Task to add the labels for. Can be left as None for a
            single task project, but is required for a task chain project
        :param revisit_affected_annotations: True to make sure that the server will
            assign a `to_revisit` status to all annotations linked to the label(s)
            that are added. False to not revisit any potentially linked annotations.
        :return: Updated Project instance with the new labels added to it
        """
        # Validate inputs
        if task is not None and task not in project.get_trainable_tasks():
            raise ValueError(
                f"The provided task {task} is not part of project {project}."
            )
        if len(project.get_trainable_tasks()) > 1 and task is None:
            raise ValueError(
                f"Project '{project}' is a task-chain project, but no target task was "
                f"specified. Please provide a valid task to perform label addition."
            )
        task_index = 0 if task is None else project.get_trainable_tasks().index(task)

        # Update the list of labels for the task
        label_list = project.get_labels_per_task()[task_index]
        existing_colors = [label["color"] for label in label_list]
        formatted_labels: List[Dict[str, Any]] = []
        for label_data in labels:
            new_color = generate_unique_label_color(existing_colors)
            if isinstance(label_data, str):
                label_dict = {
                    "name": label_data,
                    "color": new_color,
                    "group": label_data,
                }
            elif isinstance(label_data, dict):
                label_name = label_data.get("name", None)
                if label_name is None:
                    raise ValueError(
                        f"Unable to add label {label_data}: Label name not specified."
                    )
                if "color" not in label_data:
                    label_data.update({"color": new_color})
                if "group" not in label_data:
                    label_data.update({"group": label_name})
                label_dict = label_data
            else:
                raise ValueError(
                    f"Invalid input label format found for label {label_data}. Please "
                    f"provide either the label name as a string or a dictionary of "
                    f"label properties."
                )
            label_dict.update(
                {"revisit_affected_annotations": revisit_affected_annotations}
            )
            formatted_labels.append(label_dict)
            existing_colors.append(new_color)
        label_list.extend(formatted_labels)

        # Prepare data for the update request
        project.prepare_for_post()
        project_data = project.to_dict()
        task_id = project.get_trainable_tasks()[task_index].id
        task_data = next(
            (
                task
                for task in project_data["pipeline"]["tasks"]
                if task["id"] == task_id
            )
        )
        task_data["labels"] = label_list
        remove_null_fields(project_data)
        logging.info(project_data)
        response = self.session.get_rest_response(
            url=f"{self.base_url}projects/{project.id}", method="PUT", data=project_data
        )
        return ProjectRESTConverter.from_dict(response)

    def _await_project_ready(
        self, project: Project, timeout: int = 5, interval: int = 1
    ) -> None:
        """
        Await the completion of the project creation process on the Intel® Geti™ server

        :param project: Project object representing the project
        :param timeout: Time (in seconds) after which the method will time out and
            raise an error
        :param interval: Interval (in seconds) between status checks of the project
        :raises: TimeoutError if the project does not become ready after the specified
            timeout
        """
        t_start = time.time()
        error: Optional[BaseException] = None
        while time.time() - t_start < timeout:
            try:
                self.session.get_rest_response(
                    url=f"{self.base_url}projects/{project.id}/status", method="GET"
                )
                return
            except GetiRequestException as latest_error:
                time.sleep(interval)
                error = latest_error
        raise TimeoutError(
            f"Project has not become ready within the specified timeout ({timeout} "
            f"seconds)."
        ) from error

    def _get_project_detail(self, project: Union[Project, str]) -> Project:
        """
        Fetch the most recent project details from the Intel® Geti™ server

        :param project: Name of the project or Project object representing the project
            to get detailed information for.
        :return: Updated Project object
        """
        if isinstance(project, str):
            project = self.get_project_by_name(project_name=project)
            return project
        else:
            response = self.session.get_rest_response(
                url=f"{self.base_url}projects/{project.id}", method="GET"
            )
            return ProjectRESTConverter.from_dict(response)



# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import time
import warnings
from typing import List, Optional

from tqdm import TqdmWarning
from tqdm.auto import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from geti_sdk.data_models.enums.job_state import JobState
from geti_sdk.data_models.job import Job
from geti_sdk.http_session import GetiRequestException, GetiSession
from geti_sdk.rest_converters.job_rest_converter import JobRESTConverter


def restrict(value: float, min: float = 0, max: float = 1) -> float:
    """
    Restrict the input `value` to a certain range such that `min` < `value` < `max`

    :param value: Variable to restrict
    :param min: Minimum allowed value. Defaults to zero
    :param max: Maximally allowed value. Defaults to one
    :return: The value constrained to the specified range
    """
    if value < min:
        return min
    if value > max:
        return max
    return value


def get_job_by_id(
    job_id: str, session: GetiSession, workspace_id: str
) -> Optional[Job]:
    """
    Retrieve Job information from the Intel® Geti™ server.

    :param job_id: Unique ID of the job to retrieve
    :param session: GetiSession instance addressing the Intel® Geti™ platform
    :param workspace_id: ID of the workspace in which the job was created
    :return: Job instance holding the details of the job
    """
    try:
        response = session.get_rest_response(
            url=f"workspaces/{workspace_id}/jobs/{job_id}", method="GET"
        )
    except GetiRequestException as error:
        if error.status_code == 404:
            return None
        else:
            raise error
    job = JobRESTConverter.from_dict(response)
    job.workspace_id = workspace_id
    job.geti_version = session.version
    return job


def get_job_with_timeout(
    job_id: str,
    session: GetiSession,
    workspace_id: str,
    job_type: str = "training",
    timeout: int = 15,
) -> Job:
    """
    Retrieve a Job from the Intel® Geti™ server, by it's unique ID. If the job is not
    found within the specified `timeout`, a RuntimeError is raised.

    :param job_id: Unique ID of the job to retrieve
    :param session: GetiSession instance addressing the Intel® Geti™ platform
    :param workspace_id: ID of the workspace in which the job was created
    :param job_type: String representing the type of job, for instance "training" or
        "testing"
    :param timeout: Time (in seconds) after which the job retrieval will timeout
    :raises: RuntimeError if the job is not found within the specified timeout
    :return: Job instance holding the details of the job
    """
    try:
        job = get_job_by_id(job_id=job_id, session=session, workspace_id=workspace_id)
    except GetiRequestException as job_error:
        if job_error.status_code == 403:
            job = None
        else:
            raise job_error
    if job is not None:
        logging.debug(
            f"{job_type.capitalize()} job with ID {job_id} retrieved from the platform."
        )
    else:
        t_start = time.time()
        while job is None and (time.time() - t_start < timeout):
            logging.debug(
                f"{job_type.capitalize()} job status could not be retrieved from the "
                f"platform yet. Re-attempting to fetch job status. Looking for job "
                f"with ID {job_id}"
            )
            time.sleep(2)
            try:
                job = get_job_by_id(
                    session=session, job_id=job_id, workspace_id=workspace_id
                )
            except GetiRequestException as job_error:
                if job_error.status_code == 403:
                    job = None
                else:
                    raise job_error
        if job is None:
            raise RuntimeError(
                f"Unable to find the resulting {job_type} job on the Intel® Geti™ "
                f"server."
            )
    job.workspace_id = workspace_id
    job.geti_version = session.version
    return job


def monitor_jobs(
    session: GetiSession, jobs: List[Job], timeout: int = 10000, interval: int = 15
) -> List[Job]:
    """
    Monitor and print the progress of all jobs in the list `jobs`. Execution is
    halted until all jobs have either finished, failed or were cancelled.

    Progress will be reported in 15s intervals

    :param session: GetiSession instance addressing the Intel® Geti™ platform
    :param jobs: List of jobs to monitor
    :param timeout: Timeout (in seconds) after which to stop the monitoring
    :param interval: Time interval (in seconds) at which the TrainingClient polls
        the server to update the status of the jobs. Defaults to 15 seconds
    :return: List of finished (or failed) jobs with their status updated
    """
    monitoring = True
    completed_states = [
        JobState.FINISHED,
        JobState.CANCELLED,
        JobState.FAILED,
        JobState.ERROR,
    ]
    jobs_to_monitor = [job for job in jobs if job.state not in completed_states]
    logging.info(f"Monitoring progress for {len(jobs_to_monitor)} jobs...")
    outer_bars = []
    inner_bars = []
    descriptions = []
    progress_values = []
    job_steps = []
    total_job_steps = []
    finished_jobs: List[Job] = []
    jobs_with_error: List[Job] = []
    with warnings.catch_warnings(), logging_redirect_tqdm(tqdm_class=tqdm):
        warnings.filterwarnings("ignore", category=TqdmWarning)
        for index, job in enumerate(jobs_to_monitor):
            inner_description = job.current_step_message
            outer_description = f"Project `{job.metadata.project.name}` - {job.name}"
            outer_bars.append(
                tqdm(
                    total=job.total_steps,
                    desc=outer_description,
                    position=2 * index,
                    unit="step",
                    initial=job.current_step,
                    leave=True,
                    bar_format="{desc}: Step {n_fmt}/{total_fmt} |{bar}| [Total time elapsed: {elapsed}]",
                    miniters=0,
                )
            )
            inner_bar_format_string = (
                "{desc:>"
                + str(len(outer_description) + 2)
                + "}"
                + "{percentage:7.0f}% |{bar}| [{elapsed}<{remaining}, {rate_fmt}]"
            )
            inner_bar = tqdm(
                total=100,
                unit="%",
                bar_format=inner_bar_format_string,
                position=2 * index + 1,
                leave=True,
            )
            inner_bar.set_description(inner_description)
            inner_bars.append(inner_bar)
            descriptions.append(inner_description)
            progress_values.append(0)
            job_steps.append(job.current_step)
            total_job_steps.append(job.total_steps)
        try:
            t_start = time.time()
            t_elapsed = 0
            complete_count = 0
            while monitoring and t_elapsed < timeout:
                for index, job in enumerate(jobs_to_monitor):
                    if job in finished_jobs or job in jobs_with_error:
                        # Job has completed some time ago, skip further updates
                        continue

                    try:
                        job.update(session)
                    except GetiRequestException as error:
                        if error.status_code == 404:
                            logging.warning(
                                f"Job with name `{job.name}` and id `{job.id}` was not "
                                f"found on the Intel Geti instance. Monitoring is skipped "
                                f"for this job."
                            )
                        jobs_with_error.append(job)
                        complete_count += 1
                    if job.state in completed_states:
                        # Job has just completed, update progress bars to final state
                        complete_count += 1
                        finished_jobs.append(job)
                        inner_bars[index].set_description(
                            "Job completed.",
                            refresh=True,
                        )
                        inner_bars[index].update(110)
                        outer_bars[index].update(
                            total_job_steps[index] - job_steps[index]
                        )
                        continue

                    no_step_message = job.current_step_message
                    if no_step_message != descriptions[index]:
                        # Next phase of the job, reset progress bar
                        inner_bars[index].set_description(no_step_message, refresh=True)
                        inner_bars[index].reset(total=100)
                        descriptions[index] = no_step_message
                        progress_values[index] = 0
                        outer_bars[index].total = job.total_steps
                        outer_bars[index].update(job.current_step - job_steps[index])
                        job_steps[index] = job.current_step

                    incremental_progress = (
                        job.current_step_progress - progress_values[index]
                    )
                    restrict(incremental_progress, min=0, max=100)
                    inner_bars[index].update(incremental_progress)
                    progress_values[index] = job.current_step_progress
                    outer_bars[index].update(0)

                if complete_count == len(jobs_to_monitor):
                    break
                time.sleep(interval)
                t_elapsed = time.time() - t_start
        except KeyboardInterrupt:
            logging.info("Job monitoring interrupted, stopping...")
            for ib, ob in zip(inner_bars, outer_bars):
                ib.close()
                ob.close()
            return jobs
        if t_elapsed < timeout:
            logging.info("All jobs completed, monitoring stopped.")
        else:
            logging.info(
                f"Monitoring stopped after {t_elapsed:.1f} seconds due to timeout."
            )
        for ib, ob in zip(inner_bars, outer_bars):
            ib.close()
            ob.close()
    return jobs


def monitor_job(
    session: GetiSession, job: Job, timeout: int = 10000, interval: int = 15
) -> List[Job]:
    """
    Monitor and print the progress of a single `job`. Execution is
    halted until the job has either finished, failed or was cancelled.

    Progress will be reported in 15s intervals

    :param session: GetiSession instance addressing the Intel® Geti™ platform
    :param job: Job to monitor
    :param timeout: Timeout (in seconds) after which to stop the monitoring
    :param interval: Time interval (in seconds) at which the TrainingClient polls
        the server to update the status of the job. Defaults to 15 seconds
    :return: The finished (or failed) job with it's status updated
    """
    monitoring = True
    completed_states = [
        JobState.FINISHED,
        JobState.CANCELLED,
        JobState.FAILED,
        JobState.ERROR,
    ]
    try:
        logging.info(
            f"Monitoring `{job.name}` job for project `{job.metadata.project.name}`: {job.metadata.task.name}"
        )
    except AttributeError:
        logging.info(f"Monitoring `{job.name}` with id {job.id}")
    try:
        job.update(session)
    except GetiRequestException as error:
        if error.status_code == 404:
            logging.warning(
                f"Job with name `{job.name}` and id `{job.id}` was not "
                f"found on the Intel Geti instance. Monitoring is skipped "
                f"for this job."
            )
    if job.state in completed_states:
        logging.info(
            f"Job `{job.name}` has already finished with status "
            f"{str(job.state)}, monitoring stopped"
        )
        return job

    t_start = time.time()
    t_elapsed = 0

    with logging_redirect_tqdm(tqdm_class=tqdm), warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=TqdmWarning)
        try:
            previous_progress = 0
            previous_message = job.current_step_message
            current_step = job.current_step
            outer_description = f"Project `{job.metadata.project.name}` - {job.name}"
            total_steps = job.total_steps
            outer_bar = tqdm(
                total=total_steps,
                desc=outer_description,
                position=0,
                unit="step",
                initial=current_step,
                leave=True,
                bar_format="{desc}: Step {n_fmt}/{total_fmt} |{bar}| [Total time elapsed: {elapsed}]",
                miniters=0,
            )
            inner_bar_format_string = (
                "{desc:>"
                + str(len(outer_description) + 2)
                + "}"
                + "{percentage:7.0f}% |{bar}| [{elapsed}<{remaining}, {rate_fmt}]"
            )
            inner_bar = tqdm(
                total=100,
                unit="%",
                bar_format=inner_bar_format_string,
                position=1,
                leave=False,
            )
            inner_bar.set_description(previous_message)
            while monitoring and t_elapsed < timeout:
                job.update(session)
                if job.total_steps > total_steps:
                    total_steps = job.total_steps
                    outer_bar.reset(total=job.total_steps)
                    outer_bar.update(job.current_step)
                    current_step = job.current_step
                if job.state in completed_states:
                    outer_bar.update(total_steps - current_step)
                    inner_bar.update(100 - previous_progress)
                    monitoring = False
                    break
                no_step_message = job.current_step_message
                if no_step_message != previous_message:
                    # Next phase of the job, reset progress bar
                    inner_bar.set_description(f"{no_step_message}", refresh=True)
                    inner_bar.reset(total=100)
                    previous_message = no_step_message
                    previous_progress = 0
                    outer_bar.update(job.current_step - current_step)
                    current_step = job.current_step

                incremental_progress = job.current_step_progress - previous_progress
                restrict(incremental_progress, min=0, max=100)
                inner_bar.update(incremental_progress)
                outer_bar.update(0)
                previous_progress = job.current_step_progress
                time.sleep(interval)
                t_elapsed = time.time() - t_start
            inner_bar.close()
            outer_bar.close()
        except KeyboardInterrupt:
            logging.info("Job monitoring interrupted, stopping...")
            job.update(session)
            inner_bar.close()
            outer_bar.close()
            return job
        if t_elapsed < timeout:
            logging.info(
                f"Job `{job.name}` finished, monitoring stopped. Total time elapsed: "
                f"{t_elapsed:.1f} seconds"
            )
        else:
            logging.info(
                f"Monitoring stopped after {t_elapsed:.1f} seconds due to timeout. Current "
                f"job state: {job.state}"
            )
    return job



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import abc
import math
from typing import Any, Dict, List, Optional, Tuple, Union

import attr
import cv2
import numpy as np

from geti_sdk.data_models.enums import ShapeType
from geti_sdk.data_models.utils import round_to_n_digits, str_to_shape_type

# N_DIGITS_TO_ROUND_TO determines how pixel coordinates will be rounded when they are
# passed from the Intel® Geti™ REST API. The Intel® Geti™ server itself rounds some
# coordinates to 4 digits, but not all. Here we round all coordinates for internal
# consistency
N_DIGITS_TO_ROUND_TO = 0
coordinate_converter = round_to_n_digits(N_DIGITS_TO_ROUND_TO)


@attr.define(slots=False)
class Shape:
    """
    Representation of a shape in on the Intel® Geti™ platform.

    :var type: Type of the shape
    """

    type: str = attr.field(converter=str_to_shape_type)

    @abc.abstractmethod
    def to_roi(self) -> "Rectangle":
        """
        Return the bounding box containing the shape, as an instance of the Rectangle
        class.

        :return: Rectangle representing the bounding box for the shape
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_absolute_coordinates(self, parent_roi: "Rectangle") -> "Shape":
        """
        Convert the Shape to absolute coordinates, given the rectangle
        representing it's parent region of interest.

        :param parent_roi: Region of interest containing the shape
        :return: Shape converted to the coordinate system of it's parent roi
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_normalized_coordinates(
        self, image_width: int, image_height: int
    ) -> Dict[str, Any]:
        """
        Convert the Shape to a normalized coordinate system, such that all
        coordinates are represented as floats in the interval [0, 1].

        :param image_width: Width (in pixels) of the image or region with respect to
            which the shape coordinates should be normalized.
        :param image_height: Height (in pixels) of the image or region with respect to
            which the shape coordinates should be normalized
        :return: Dictionary representing the Shape object with it's coordinates
            normalized with respect to image_width and image_height
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def area(self) -> float:
        """
        Return the area of the shape, given in number of pixels.

        :return: Area of the shape, in pixels
        """
        raise NotImplementedError


@attr.define(slots=False)
class Rectangle(Shape):
    """
    Representation of a Rectangle on the Intel® Geti™ platform, as used in the
    /annotations REST endpoints.

    NOTE: All coordinates and dimensions are given in pixels

    :var x: X coordinate of the left side of the rectangle
    :var y: Y coordinate of the top of the rectangle
    :var width: Width of the rectangle
    :var height: Height of the rectangle
    """

    x: int = attr.field(converter=coordinate_converter)
    y: int = attr.field(converter=coordinate_converter)
    width: int = attr.field(converter=coordinate_converter)
    height: int = attr.field(converter=coordinate_converter)
    type: str = attr.field(
        converter=str_to_shape_type, default=ShapeType.RECTANGLE, kw_only=True
    )

    def to_normalized_coordinates(
        self, image_width: int, image_height: int
    ) -> Dict[str, float]:
        """
        Get the normalized coordinates of the rectangle, with respect to the image
        with dimensions `image_width` x `image_height`.

        :param image_width: Width of the image to which the coordinates should be
            normalized
        :param image_height: Height of the image to which the coordinates should be
            normalized
        :return: Dictionary containing the rectangle, represented in normalized
            coordinates
        """
        return dict(
            x=self.x / image_width,
            y=self.y / image_height,
            width=self.width / image_width,
            height=self.height / image_height,
            type=str(self.type),
        )

    def is_full_box(self, image_width: int, image_height: int) -> bool:
        """
        Return True if this Rectangle represents a box containing the full image.

        :param image_width: Width of the image to check for
        :param image_height: Height of the image to check for
        :return: True if the Rectangle encompasses the full image, False otherwise
        """
        return (
            self.x == 0
            and self.y == 0
            and int(self.width) == image_width
            and int(self.height) == image_height
        )

    def to_roi(self) -> "Rectangle":
        """
        Return the bounding box containing the shape, as an instance of the Rectangle
        class.

        :return: Rectangle representing the bounding box for the shape
        """
        return self

    def to_absolute_coordinates(self, parent_roi: "Rectangle") -> "Rectangle":
        """
        Convert the Rectangle to absolute coordinates, given the rectangle
        representing it's parent region of interest.

        :param parent_roi: Region of interest containing the rectangle
        :return: Rectangle converted to the coordinate system of it's parent
        """
        x_min = parent_roi.x + self.x
        y_min = parent_roi.y + self.y
        return Rectangle(x=x_min, y=y_min, width=self.width, height=self.height)

    @property
    def area(self) -> float:
        """
        Return the area of the Rectangle, in pixels.

        :return:
        """
        return self.width * self.height

    @property
    def x_max(self) -> int:
        """
        Return the value of the maximal x-coordinate that the Rectangle instance
        touches.

        :return: Maximum x-coordinate for the rectangle
        """
        return self.x + self.width

    @property
    def y_max(self) -> int:
        """
        Return the value of the maximal y-coordinate that the Rectangle instance
        touches.

        :return: Maximum y-coordinate for the rectangle
        """
        return self.y + self.height

    @classmethod
    def generate_full_box(cls, image_width: int, image_height: int) -> "Rectangle":
        """
        Return a rectangle that fully encapsulates the image.

        :param image_width: Width of the image to which the rectangle applies (in pixels)
        :param image_height: Height of the image to which the rectangle applies (in pixels)
        :return: Rectangle: A rectangle that fully encapsulates the image.
        """
        return cls(x=0, y=0, width=image_width, height=image_height)


@attr.define(slots=False)
class Ellipse(Shape):
    """
    Representation of an Ellipse on the Intel® Geti™ platform, as used in the
    /annotations REST endpoints.

    NOTE: All coordinates and dimensions are given in pixels

    :var x: Lowest x coordinate of the ellipse
    :var y: Lowest y coordinate of the ellipse
    :var width: Width of the ellipse
    :var height: Height of the ellipse
    """

    x: int = attr.field(converter=coordinate_converter)
    y: int = attr.field(converter=coordinate_converter)
    width: int = attr.field(converter=coordinate_converter)
    height: int = attr.field(converter=coordinate_converter)
    type: str = attr.field(
        converter=str_to_shape_type, default=ShapeType.ELLIPSE, kw_only=True
    )

    def to_roi(self) -> "Rectangle":
        """
        Return the bounding box containing the Ellipse, as an instance of the Rectangle
        class.

        :return: Rectangle representing the bounding box for the ellipse
        """
        return Rectangle(x=self.x, y=self.y, width=self.width, height=self.height)

    def to_absolute_coordinates(self, parent_roi: Rectangle) -> "Ellipse":
        """
        Convert the Ellipse to absolute coordinates, given the rectangle
        representing it's parent region of interest.

        :param parent_roi: Region of interest containing the ellipse
        :return: Ellipse converted to the coordinate system of it's parent
        """
        x_min = parent_roi.x + self.x
        y_min = parent_roi.y + self.y
        return Ellipse(x=x_min, y=y_min, width=self.width, height=self.height)

    def to_normalized_coordinates(
        self, image_width: int, image_height: int
    ) -> Dict[str, float]:
        """
        Get the normalized coordinates of the ellipse, with respect to the image
        with dimensions `image_width` x `image_height`.

        :param image_width: Width of the image to which the coordinates should be
            normalized
        :param image_height: Height of the image to which the coordinates should be
            normalized
        :return: Dictionary containing the ellipse, represented in normalized
            coordinates
        """
        return dict(
            x=self.x / image_width,
            y=self.y / image_height,
            width=self.width / image_width,
            height=self.height / image_height,
            type=str(self.type),
        )

    def get_center_point(self) -> Tuple[int, int]:
        """
        Return the coordinates of the center of the ellipse.

        :return: Tuple of integers representing the coordinates of the center of the ellipse
        """
        return self.x + self.width // 2, self.y + self.height // 2

    @property
    def area(self) -> float:
        """
        Return the area of the Ellipse, in pixels.

        :return:
        """
        return self.width * self.height * math.pi

    @property
    def x_max(self) -> int:
        """
        Return the value of the maximal x-coordinate that the Ellipse instance
        touches.

        :return: Maximum x-coordinate for the ellipse
        """
        return self.x + self.width

    @property
    def y_max(self) -> int:
        """
        Return the value of the maximal y-coordinate that the Ellipse instance
        touches.

        :return: Maximum y-coordinate for the ellipse
        """
        return self.y + self.height


@attr.define()
class Point:
    """
    Representation of a point on a 2D coordinate system. Used to define Polygons on
    the Intel® Geti™ platform.

    NOTE: All coordinates are defined in pixels

    :var x: X coordinate of the point
    :var y: Y coordinate of the point
    """

    x: int = attr.field(converter=coordinate_converter)
    y: int = attr.field(converter=coordinate_converter)

    def as_int_tuple(self) -> Tuple[int, int]:
        """
        Return the coordinates of the point as a tuple of integers.

        :return: Tuple of integers representing the coordinates of the point
        """
        return int(self.x), int(self.y)


@attr.define(slots=False)
class Polygon(Shape):
    """
    Representation of a polygon on the Intel® Geti™ platform, as used in the
    /annotations REST endpoints.

    :var points: List of Points that make up the polygon
    """

    points: List[Point]
    type: str = attr.field(
        converter=str_to_shape_type, default=ShapeType.POLYGON, kw_only=True
    )

    def __attrs_post_init__(self):
        """
        Initialize private attributes.
        """
        self._contour: Optional[np.ndarray] = None
        self._x_max: Optional[int] = None
        self._y_max: Optional[int] = None

    def points_as_contour(self) -> np.ndarray:
        """
        Return the list of points for this Polygon as a numpy array representing
        contour points that can be plotted by openCV's drawContours function.

        NOTE: the contour is cached, to avoid going over the list of points twice. If
        the polygon is modified the cache has to be cleared. It is recommended to
        create a new polygon instead

        :return: Numpy array containing the contour
        """
        if self._contour is None:
            self._contour = np.array(
                [(int(point.x), int(point.y)) for point in self.points]
            )
        return self._contour

    def to_roi(self) -> "Rectangle":
        """
        Return the bounding box containing the Polygon, as an instance of the Rectangle
        class.

        :return: Rectangle representing the bounding box for the polygon
        """
        points_array = np.array([(point.x, point.y) for point in self.points])
        min_xy = points_array.min(axis=0)
        max_xy = points_array.max(axis=0)

        return Rectangle(
            x=min_xy[0],
            y=min_xy[1],
            width=max_xy[0] - min_xy[0],
            height=max_xy[1] - min_xy[1],
        )

    def to_absolute_coordinates(self, parent_roi: Rectangle) -> "Polygon":
        """
        Convert the Polygon to absolute coordinates, given the rectangle
        representing it's parent region of interest.

        :param parent_roi: Region of interest containing the polygon
        :return: Polygon converted to the coordinate system of it's parent ROI
        """
        absolute_points = [
            Point(x=parent_roi.x + point.x, y=parent_roi.y + point.y)
            for point in self.points
        ]
        return Polygon(points=absolute_points)

    def to_normalized_coordinates(
        self, image_width: int, image_height: int
    ) -> Dict[str, Union[List[Dict[str, float]], str]]:
        """
        Get the normalized coordinates of the polygon, with respect to the image
        with dimensions `image_width` x `image_height`.

        :param image_width: Width of the image to which the coordinates should be
            normalized
        :param image_height: Height of the image to which the coordinates should be
            normalized
        :return: Dictionary containing the polygon, represented in normalized
            coordinates
        """
        normalized_points: List[Dict[str, float]] = []
        for point in self.points:
            normalized_points.append(
                {"x": point.x / image_width, "y": point.y / image_height}
            )
        return dict(points=normalized_points, type=str(self.type))

    @property
    def area(self) -> float:
        """
        Return the area of the Polygon, in pixels.

        :return: area of the polygon
        """
        return cv2.contourArea(self.points_as_contour())

    def _calculate_xy_max(self):
        """
        Calculate the maximum x and y coordinates that the Polyon touches, in pixels.
        """
        coord_maxes = self.points_as_contour().max(axis=0)
        self._x_max, self._y_max = coord_maxes[0], coord_maxes[1]

    @property
    def x_max(self) -> int:
        """
        Return the maximum x-coordinate of the Polygon, in pixels

        :return: largest x-coordinate that the polygon touches
        """
        if self._x_max is None:
            self._calculate_xy_max()
        return self._x_max

    @property
    def y_max(self) -> int:
        """
        Return the maximum y-coordinate of the Polygon, in pixels

        :return: largest y-coordinate that the polygon touches
        """
        if self._y_max is None:
            self._calculate_xy_max()
        return self._y_max

    def fit_rotated_rectangle(self) -> "RotatedRectangle":
        """
        Fit a RotatedRectangle object around the Polygon, such that the area spanned
        by the rectangle is minimal.

        :return: RotatedRectangle object with minimal area, which encloses the Polygon
        """
        center, (width, height), angle = cv2.minAreaRect(self.points_as_contour())
        return RotatedRectangle(
            angle=angle, x=center[0], y=center[1], width=width, height=height
        )


@attr.define(slots=False)
class RotatedRectangle(Shape):
    """
    Representation of a RotatedRectangle on the Intel® Geti™ platform, as used in the
    /annotations REST endpoints.

    NOTE: All coordinates and dimensions are specified in pixels

    :var angle: angle, in degrees, under which the rectangle is defined.
    :var x: X coordinate of the center of the rectangle
    :var y: Y coordinate of the center of the rectangle
    :var width: Width of the rectangle
    :var height: Height of the rectangle
    """

    angle: float = attr.field(converter=round_to_n_digits(n=4))
    x: int = attr.field(converter=coordinate_converter)
    y: int = attr.field(converter=coordinate_converter)
    width: int = attr.field(converter=coordinate_converter)
    height: int = attr.field(converter=coordinate_converter)
    type: str = attr.field(
        converter=str_to_shape_type, default=ShapeType.ROTATED_RECTANGLE, kw_only=True
    )

    @property
    def _angle_x_radian(self) -> float:
        """
        Return the angle to the x-axis, in radians.

        :return: angle to the x-axis in radians
        """
        return (2 * math.pi / 360) * self.angle

    @property
    def x_max(self) -> float:
        """
        Return the value of the maximal x-coordinate that the rotated rectangle touches.

        :return: Maximum x-coordinate for the rotated rectangle
        """
        return (
            self.x
            + 0.5 * self.width * math.cos(self._angle_x_radian)
            + 0.5 * self.height * math.sin(self._angle_x_radian)
        )

    @property
    def x_min(self) -> float:
        """
        Return the value of the minimal x-coordinate that the rotated rectangle touches.

        :return: Minimum x-coordinate for the rotated rectangle
        """
        return (
            self.x
            - 0.5 * self.width * math.cos(self._angle_x_radian)
            - 0.5 * self.height * math.sin(self._angle_x_radian)
        )

    @property
    def y_max(self) -> float:
        """
        Return the value of the maximal y-coordinate that the rotated rectangle touches.

        :return: Maximum y-coordinate for the rotated rectangle
        """
        return (
            self.y
            + 0.5 * self.width * math.sin(self._angle_x_radian)
            + 0.5 * self.height * math.cos(self._angle_x_radian)
        )

    @property
    def y_min(self):
        """
        Return the value of the minimal y-coordinate that the rotated rectangle touches.

        :return: Minimum y-coordinate for the rotated rectangle
        """
        return (
            self.y
            - 0.5 * self.width * math.sin(self._angle_x_radian)
            - 0.5 * self.height * math.cos(self._angle_x_radian)
        )

    @classmethod
    def from_polygon(cls, polygon: Polygon) -> "RotatedRectangle":
        """
        Create a :py:class`~geti_sdk.data_models.shapes.RotatedRectangle` from
        the Polygon entity passed.

        NOTE: The Polygon MUST consist of 4 points, otherwise a ValueError is raised

        :param polygon: Polygon entity to convert from
        :return: Rectangle instance created according to the polygon object
        """
        if len(polygon.points) != 4:
            raise ValueError(
                f"Unable to convert polygon {polygon} to RotatedRectangle. A rotated "
                f"rectangle must have exactly 4 points."
            )

        x_coords = [point.x for point in polygon.points]
        y_coords = [point.y for point in polygon.points]

        x_min, x_max = min(x_coords), max(x_coords)
        y_min, y_max = min(y_coords), max(y_coords)

        x_center = x_min + 0.5 * (x_max - abs(x_min))
        y_center = y_min + 0.5 * (y_max - abs(y_min))

        if len(x_coords) > len(set(x_coords)) or len(y_coords) > len(set(y_coords)):
            # In this case there are points sharing the same x or y value, which means
            # that we have an angle of 0 degrees
            width = x_max - x_min
            height = y_max - y_min
            alpha = 0
        else:
            x_min_coord = Point(x=x_min, y=y_coords[x_coords.index(x_min)])
            x_max_coord = Point(x=x_max, y=y_coords[x_coords.index(x_max)])
            y_min_coord = Point(x=x_coords[y_coords.index(y_min)], y=y_min)

            # Calculate the angle to the x-axis, alpha
            perpendicular_side = x_max_coord.y - y_min_coord.y
            base_side = x_max_coord.x - y_min_coord.x
            hypotenuse = math.sqrt(perpendicular_side**2 + base_side**2)

            alpha = math.asin(perpendicular_side / hypotenuse)
            alpha = alpha * 360 / (2 * math.pi)

            width = hypotenuse

            # Calculate height
            perpendicular_side_2 = x_min_coord.y - y_min_coord.y
            base_side_2 = x_min_coord.x - y_min_coord.x
            height = math.sqrt(perpendicular_side_2**2 + base_side_2**2)

        return cls(
            x=int(x_center),
            y=int(y_center),
            width=int(width),
            height=int(height),
            angle=alpha,
        )

    def to_roi(self) -> Rectangle:
        """
        Return the bounding box containing the RotatedRectangle, as an instance of
        the Rectangle class.

        :return: Rectangle representing the bounding box for the rotated_rectangle
        """
        return Rectangle(
            x=self.x,
            y=self.y,
            width=int(self.x_max - self.x_min),
            height=int(self.y_max - self.y_min),
        )

    def to_absolute_coordinates(self, parent_roi: Rectangle) -> "RotatedRectangle":
        """
        Convert the RotatedRectangle to absolute coordinates, given the rectangle
        representing it's parent region of interest.

        :param parent_roi: Region of interest containing the rotated rectangle
        :return: RotatedRectangle converted to the coordinate system of it's parent ROI
        """
        x = parent_roi.x + self.x
        y = parent_roi.y + self.y

        width = self.width
        height = self.height
        return RotatedRectangle(x=x, y=y, width=width, height=height, angle=self.angle)

    def to_polygon(self) -> Polygon:
        """
        Convert the RotatedRectangle instance to a Polygon consisting of 4 points.

        :return: Polygon object corresponding to the RotatedRectangle instance
        """
        y_0 = (
            self.y
            - 0.5 * self.width * math.sin(self._angle_x_radian)
            + 0.5 * self.height * math.cos(self._angle_x_radian)
        )

        x_1 = (
            self.x
            - 0.5 * self.width * math.cos(self._angle_x_radian)
            + 0.5 * self.height * math.sin(self._angle_x_radian)
        )

        y_2 = (
            self.y
            + 0.5 * self.width * math.sin(self._angle_x_radian)
            - 0.5 * self.height * math.cos(self._angle_x_radian)
        )

        x_3 = (
            self.x
            + 0.5 * self.width * math.cos(self._angle_x_radian)
            - 0.5 * self.height * math.sin(self._angle_x_radian)
        )

        point0 = Point(x=int(self.x_min), y=int(y_0))
        point1 = Point(x=int(x_1), y=int(self.y_min))
        point2 = Point(x=int(self.x_max), y=int(y_2))
        point3 = Point(x=int(x_3), y=int(self.y_max))

        return Polygon(points=[point0, point1, point2, point3])

    def to_normalized_coordinates(
        self, image_width: int, image_height: int
    ) -> Dict[str, Union[float, str]]:
        """
        Get the normalized coordinates of the rotated rectangle, with respect to the
        image with dimensions `image_width` x `image_height`.

        :param image_width: Width of the image to which the coordinates should be
            normalized
        :param image_height: Height of the image to which the coordinates should be
            normalized
        :return: Dictionary containing the rotated rectangle, represented in normalized
            coordinates
        """
        return dict(
            angle=self.angle,
            x=self.x / image_width,
            y=self.y / image_height,
            width=self.width / image_width,
            height=self.height / image_height,
            type=str(self.type),
        )

    @property
    def area(self) -> float:
        """
        Return the area of the RotatedRectangle, in pixels

        :return: area enclosed by the rotated rectangle in pixels
        """
        return self.width * self.height



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
import logging
import os
import tempfile
import time
import zipfile
from typing import List, Optional, Sequence, Union

from geti_sdk.data_models import Project
from geti_sdk.data_models.code_deployment_info import (
    CodeDeploymentInformation,
    DeploymentModelIdentifier,
)
from geti_sdk.data_models.enums import DeploymentState, OptimizationType
from geti_sdk.data_models.model import Model, OptimizedModel
from geti_sdk.deployment import Deployment
from geti_sdk.http_session import GetiSession
from geti_sdk.rest_clients.configuration_client import ConfigurationClient
from geti_sdk.rest_clients.model_client import ModelClient
from geti_sdk.rest_clients.prediction_client import PredictionClient
from geti_sdk.utils import deserialize_dictionary, get_supported_algorithms


class DeploymentClient:
    """
    Class to manage model deployment for a certain Intel® Geti™ project.
    """

    def __init__(self, workspace_id: str, project: Project, session: GetiSession):
        self.session = session
        self.project = project
        self.workspace_id = workspace_id
        self.base_url = f"workspaces/{workspace_id}/projects/{project.id}"
        self.supported_algos = get_supported_algorithms(
            rest_session=session, project=project, workspace_id=workspace_id
        )

        self._model_client = ModelClient(
            workspace_id=workspace_id, project=project, session=session
        )
        self._prediction_client = PredictionClient(
            workspace_id=workspace_id, project=project, session=session
        )
        self._configuration_client = ConfigurationClient(
            workspace_id=workspace_id, project=project, session=session
        )

    @property
    def code_deployment_url(self) -> str:
        """
        Return the base URL for the code deployment group of endpoints

        :return: URL for the code deployment endpoints for the Intel® Geti™ project
        """
        return self.base_url + "/code_deployments"

    @property
    def ready_to_deploy(self) -> bool:
        """
        Return True when the project is ready for deployment, False otherwise.

        A project is ready for deployment when it contains at least one trained model
        for each task.

        :return: True when the project is ready for deployment, False otherwise
        """
        return self._prediction_client.ready_to_predict and all(
            model is not None for model in self._model_client.get_all_active_models()
        )

    def _request_deployment(
        self, model_identifiers: Sequence[DeploymentModelIdentifier]
    ) -> CodeDeploymentInformation:
        """
        Request the creation of a code deployment for the project, using the models
        with identifiers as passed in `models`

        :param model_identifiers: List of identifiers for the models to use in the
            deployment
        :return: CodeDeploymentInformation object, holding all available information
            regarding the state of the deployment creation process on the Intel® Geti™
            server.
        """
        model_id_data = [model.to_dict() for model in model_identifiers]
        response = self.session.get_rest_response(
            url=self.code_deployment_url + ":prepare",
            method="POST",
            data={"models": model_id_data},
        )
        logging.info(f"Deployment for project '{self.project.name}' started.")
        return deserialize_dictionary(response, CodeDeploymentInformation)

    def _get_deployment_status(self, deployment_id: str) -> CodeDeploymentInformation:
        """
        Return the actual status of a code deployment creation process on the Intel®
        Geti™ server.

        :param deployment_id: Unique identifier of the deployment to get the status for
        :return: CodeDeploymentInformation instance holding the current state of the
            deployment creation process
        """
        response = self.session.get_rest_response(
            url=self.code_deployment_url + f"/{deployment_id}", method="GET"
        )
        return deserialize_dictionary(response, CodeDeploymentInformation)

    def _fetch_deployment(self, deployment_id: str) -> Deployment:
        """
        Download and return the deployment identified by `deployment_id`. If the
        deployment is not ready to be downloaded yet, this method will raise an error.

        :param deployment_id: Unique identifier of the deployment to download
        :return: `Deployment` object containing the deployment
        """
        status = self._get_deployment_status(deployment_id=deployment_id)
        if status.state != DeploymentState.DONE:
            raise ValueError(
                f"Deployment with ID '{deployment_id}' is not ready to be retrieved "
                f"(yet). The current state is '{status.state}'. Deployment can only be "
                f"retrieved when it is in state 'DONE'. "
            )
        deployment_tempdir = tempfile.mkdtemp()
        zipfile_path = os.path.join(deployment_tempdir, "deployment.zip")
        response = self.session.get_rest_response(
            url=self.code_deployment_url + f"/{deployment_id}/download",
            method="GET",
            contenttype="zip",
        )
        logging.info("Downloading project deployment archive...")
        with open(zipfile_path, "wb") as f:
            f.write(response.content)
        with zipfile.ZipFile(zipfile_path, "r") as zipped_deployment:
            zipped_deployment.extractall(deployment_tempdir)
        logging.info(
            f"Deployment for project '{self.project.name}' downloaded and extracted "
            f"successfully."
        )
        if os.path.exists(zipfile_path):
            os.remove(zipfile_path)
        deployment = Deployment.from_folder(
            path_to_folder=os.path.join(deployment_tempdir, "deployment")
        )
        deployment._path_to_temp_resources = deployment_tempdir
        deployment._requires_resource_cleanup = True
        return deployment

    def deploy_project(
        self,
        output_folder: Optional[Union[str, os.PathLike]] = None,
        models: Optional[Sequence[Union[Model, OptimizedModel]]] = None,
        enable_explainable_ai: bool = False,
        prepare_for_ovms: bool = False,
    ) -> Deployment:
        """
        Deploy a project by creating a Deployment instance. The Deployment contains
        the optimized active models for each task in the project, and can be loaded
        with OpenVINO to run inference locally.

        The `models` parameter can be used in two ways:

          - If `models` is left as None this method will create a deployment containing
            the current active model for each task in the project.

          - If a list of models is passed, then it should always contain a number of
            Models less than or equal to the amount of tasks in the project. For
            instance, for a task-chain project with two tasks a list of at most two
            models could be passed (one for each task).
            If in this case a list of only one model is passed, that model will be
            used for its corresponding task and the method will resort to using the
            active model for the other task in the project.

        Optionally, configuration files for deploying models to OpenVINO Model Server
        (OVMS) can be generated. If you want this, make sure to pass
        `prepare_for_ovms=True` as input parameter. In this case, you MUST specify an
        `output_folder` to save the deployment and OVMS configuration files.
        Note that this does not change the resulting deployment in itself, it can
        still be used locally as well.

        :param output_folder: Path to a folder on local disk to which the Deployment
            should be downloaded. If no path is specified, the deployment will not be
            saved to disk directly. Note that it is always possible to save the
            deployment once it has been created, using the `deployment.save` method.
        :param models: Optional list of models to use in the deployment. If no list is
            passed, this method will create a deployment using the currently active
            model for each task in the project.
        :param enable_explainable_ai: True to include an Explainable AI head in
            the deployment. This will add an Explainable AI head to the model for each
            task in the project, allowing for the generation of saliency maps.
        :param prepare_for_ovms: True to prepare the deployment to be hosted on a
            OpenVINO model server (OVMS). Passing True will create OVMS configuration
            files for the model(s) in the project and instructions with sample code
            showing on how to launch an OVMS container including the deployed models.
        :return: Deployment for the project
        """
        if not self.ready_to_deploy:
            raise ValueError(
                f"Project '{self.project.name}' is not ready for deployment, please "
                f"make sure that each task in the project has at least one model "
                f"trained before deploying the project."
            )
        if prepare_for_ovms and output_folder is None:
            raise ValueError(
                "An `output_folder` must be specified when setting "
                "`prepare_for_ovms=True`."
            )
        deployment = self._backend_deploy_project(
            output_folder=output_folder,
            models=models,
            require_xai=enable_explainable_ai,
        )
        if prepare_for_ovms:
            deployment.generate_ovms_config(output_folder=output_folder)
        return deployment

    def _backend_deploy_project(
        self,
        output_folder: Optional[Union[str, os.PathLike]] = None,
        models: Optional[Sequence[Union[Model, OptimizedModel]]] = None,
        require_xai: bool = False,
    ) -> Deployment:
        """
        Create a deployment for a project, using the /code_deployment endpoint from the
        backend.

        :param output_folder: Path to a folder on local disk to which the Deployment
            should be downloaded. If no path is specified, the deployment will not be
            saved to disk directly. Note that it is always possible to save the
            deployment once it has been created, using the `deployment.save` method.
        :param models: Optional list of models to use in the deployment. If no list is
            passed, this method will create a deployment using the currently active
            model for each task in the project.
        :param require_xai: True to include an Explainable AI head in the deployment.
        :return: Deployment for the project
        """
        # Get the models to deploy
        tasks = self.project.get_trainable_tasks()
        if models is None:
            models = self._model_client.get_all_active_models()

        tasks_with_model = [
            self._model_client.get_task_for_model(model) for model in models
        ]

        if len(models) < len(tasks):
            tasks_without_model = [
                task for task in tasks if task not in tasks_with_model
            ]
            active_models = [
                self._model_client.get_active_model_for_task(task)
                for task in tasks_without_model
            ]

            models = models + active_models

            model_id_to_task_id = {
                model.id: task.id
                for model, task in zip(models, tasks_with_model + tasks_without_model)
            }
        else:
            model_id_to_task_id = {
                model.id: task.id for model, task in zip(models, tasks_with_model)
            }

        # Fetch the optimized model for each model, if it is not optimized already
        optimized_models: List[OptimizedModel] = []
        for model in models:
            if not isinstance(model, OptimizedModel):
                current_optim_models = model.optimized_models
                optimization_types = set(
                    (op_model.optimization_type for op_model in current_optim_models)
                )
                preferred_model = None
                for optimization_type in OptimizationType:
                    # The order of fields in OptimizationType gives us priority order for optimization types
                    if optimization_type not in optimization_types or (
                        optimization_type
                        in (OptimizationType.ONNX, OptimizationType.NONE)
                    ):
                        continue
                    for op_model in current_optim_models:
                        if (
                            op_model.optimization_type == optimization_type
                            and require_xai == op_model.has_xai_head
                        ):
                            preferred_model = op_model
                            break
                if preferred_model is None:
                    raise ValueError(
                        f"Could not find an optimized model for model '{model.name}' "
                        f"with optimization type '{optimization_type}' and explainable ai "
                        + ("enabled." if require_xai else "disabled.")
                    )
                optimized_models.append(preferred_model)

                # Update the task id for the model
                task_id_for_model = model_id_to_task_id[model.id]
                model_id_to_task_id.update({preferred_model.id: task_id_for_model})
            else:
                optimized_models.append(model)

        model_identifiers = [
            DeploymentModelIdentifier.from_model(model) for model in optimized_models
        ]
        # Make sure models are sorted according to task order
        sorted_optimized_models: List[OptimizedModel] = []
        for task in tasks:
            for model in optimized_models:
                model_task_id = model_id_to_task_id[model.id]
                if model_task_id == task.id:
                    sorted_optimized_models.append(model)
                    break

        # Retrieve hyper parameters for the models
        hyper_parameters = [
            self._configuration_client.get_for_model(
                model_id=model.id, task_id=model_id_to_task_id[model.id]
            )
            for model in sorted_optimized_models
        ]

        # Make the request to prepare the deployment
        code_deployment = self._request_deployment(model_identifiers=model_identifiers)

        # Wait for the deployment to become available
        stop_polling = False
        t_start = time.time()
        timeout = 600
        logging.info("Waiting for the deployment to be created...")
        while not stop_polling and time.time() - t_start < timeout:
            time.sleep(1)
            code_deployment = self._get_deployment_status(code_deployment.id)
            if code_deployment.state == DeploymentState.DONE:
                stop_polling = True
            elif code_deployment.state == DeploymentState.FAILED:
                raise ValueError(
                    f"The Intel® Geti™ server failed to create deployment for "
                    f"project '{self.project.name}'."
                )

        if code_deployment.state != DeploymentState.DONE:
            raise ValueError(
                f"The Intel® Geti™ server failed to create deployment for "
                f"project '{self.project.name}' within 10 minutes. Deployment "
                f"creation timed out, the current state of the deployment is"
                f" `{code_deployment.state}` with a progress of "
                f"{code_deployment.progress:.1f}%"
            )

        # Fetch the deployment package
        deployment = self._fetch_deployment(deployment_id=code_deployment.id)

        # Attach the appropriate hyper parameters and model_group_id to the deployed
        # models
        for index, model in enumerate(deployment.models):
            model.hyper_parameters = hyper_parameters[index]
            model.model_group_id = sorted_optimized_models[index].model_group_id

        # Save the deployment, if needed
        if output_folder is not None:
            deployment.save(path_to_folder=output_folder)
        return deployment



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from datetime import datetime
from typing import Any, ClassVar, Dict, List, Optional

import attr
import numpy as np

from geti_sdk.data_models.annotation_scene import AnnotationScene
from geti_sdk.data_models.annotations import Annotation
from geti_sdk.data_models.enums import AnnotationKind
from geti_sdk.data_models.label import Label
from geti_sdk.data_models.media import MediaInformation
from geti_sdk.data_models.utils import (
    deidentify,
    str_to_annotation_kind,
    str_to_datetime,
)
from geti_sdk.http_session import GetiSession


@attr.define
class ResultMedium:
    """
    Representation of a single result medium in Intel® Geti™.

    :var name: Name of the result medium option
    :var type: Type of the result medium represented by this object
    :var url: URL at which the full result medium can be downloaded
    :var id: Unique database ID assigned to the result medium
    :var label_id: Unique database ID of the label referenced by this result medium
    """

    _identifier_fields: ClassVar[str] = ["id", "data", "label_id", "url"]

    name: str
    type: Optional[str] = None
    url: Optional[str] = None
    label_id: Optional[str] = None
    id: Optional[str] = None
    roi: Optional[Dict[str, Any]] = None
    label_name: Optional[str] = attr.field(init=False, default=None)
    data: Optional[bytes] = attr.field(default=None, repr=False, init=False)

    def resolve_label_name(self, labels: List[Label]):
        """
        Add the label name to the result medium, by matching the label_id to a list of
        Labels.

        :param labels: List of Labels to get the name from
        """
        self.label_name = next(
            (label.name for label in labels if label.id == self.label_id), None
        )
        # Label id is not defined for anomaly classification, it will always return
        # the saliency map for the 'Anomalous' label
        if self.name.lower() == "anomaly map" and self.label_name is None:
            self.label_name = "Anomalous"

    def get_data(self, session: GetiSession) -> bytes:
        """
        Download the data belonging to this ResultMedium object.

        :param session: REST session to the Intel® Geti™ server from which this ResultMedium
            was generated
        :return: bytes object holding the data, if any is found
        """
        if self.data is None:
            if self.url is not None:
                response = session.get_rest_response(
                    url=self.url, method="GET", contenttype="jpeg"
                )
                if response.status_code == 200:
                    self.data = response.content
                else:
                    raise ValueError(
                        f"Unable to retrieve data for result medium {self}, received "
                        f"response {response} from Intel® Geti™ server."
                    )
        return self.data

    @property
    def friendly_name(self) -> str:
        """
        Return a human readable name with which the result medium can be identified.

        :return: friendly name for the result medium
        """
        return (
            self.name + "_" + self.label_name
            if self.label_name is not None
            else self.name
        )


@attr.define
class Prediction(AnnotationScene):
    """
    Representation of the model predictions for a certain media entity in Intel® Geti™.

    :var annotations: List of predictions belonging to the media entity
    :var id: unique database ID of the Prediction in Intel® Geti™
    :var kind: Kind of prediction (Annotation or Prediction)
    :var media_identifier: Identifier of the media entity to which this Prediction
        applies
    :var modified: Date and time at which this Prediction was last modified
    :var maps: List of additional result media belonging to this prediction
    :var feature_vector: Optional feature vector (produced by the model) for the image
        to which the prediction relates
    :var active_score: Optional active score (produced by the model) for the image
        to which the prediction relates
    """

    kind: str = attr.field(
        converter=str_to_annotation_kind,
        default=AnnotationKind.PREDICTION,
        kw_only=True,
    )
    maps: List[ResultMedium] = attr.field(factory=list, kw_only=True)
    feature_vector: Optional[np.ndarray] = attr.field(
        kw_only=True, default=None, repr=False
    )
    created: Optional[str] = attr.field(converter=str_to_datetime, default=None)

    def resolve_labels_for_result_media(self, labels: List[Label]) -> None:
        """
        Resolve the label names for all result media available with this Prediction.

        :param labels: List of Labels for the project, from which the names are taken
        """
        for map_ in self.maps:
            map_.resolve_label_name(labels=labels)

    def deidentify(self) -> None:
        """
        Remove all unique database ID's from the prediction and the entities it
        contains.
        """
        deidentify(self)
        self.media_identifier = None
        for annotation in self.annotations:
            annotation.deidentify()
        for map_ in self.maps:
            deidentify(map_)

    @property
    def has_result_media(self) -> bool:
        """
        Return True if this Prediction has result media belonging to it, False
        otherwise.

        :return: True if there are result media belonging to the prediction
        """
        return len(self.maps) > 0

    def get_result_media_data(self, session: GetiSession) -> List[ResultMedium]:
        """
        Download the data for all result media belonging to this prediction.

        :param session: REST session to the Intel® Geti™ server from which this Prediction
            was generated
        :return: List of result media, that have their data downloaded from the cluster
        """
        result: List[ResultMedium] = []
        for medium in self.maps:
            medium.get_data(session=session)
            result.append(medium)
        return result

    def as_mask(
        self,
        media_information: MediaInformation,
        probability_threshold: Optional[float] = None,
    ) -> np.ndarray:
        """
        Convert the shapes in the prediction to a mask that can be overlayed on an
        image.

        :param media_information: MediaInformation object containing the width and
            heigth of the image for which the mask should be generated.
        :param probability_threshold: Threshold (between 0 and 1) for the probability.
            Shapes that are predicted with a probability below this threshold will
            not be plotted in the mask. If left as None (the default), all predictions
            will be shown.
        :return: np.ndarray holding the mask representation of the prediction
        """
        image_width = media_information.width
        image_height = media_information.height
        mask = np.zeros((image_height, image_width, 3), dtype=np.uint8)

        for annotation in self.annotations:
            max_prob_label_index = int(
                np.argmax([label.probability for label in annotation.labels])
            )
            max_prob_label = annotation.labels[max_prob_label_index]
            if probability_threshold is not None:
                if max_prob_label.probability < probability_threshold:
                    continue
            color = max_prob_label.color_tuple
            line_thickness = 3
            shape = annotation.shape
            mask = self._add_shape_to_mask(
                shape=shape,
                mask=mask,
                labels=annotation.labels,
                color=color,
                line_thickness=line_thickness,
            )
        return mask

    def filter_by_confidence(self, confidence_threshold: float) -> "Prediction":
        """
        Return a new Prediction instance containing only those predicted annotations
        that have a confidence higher than `confidence_threshold`.

        :param confidence_threshold: Float between 0 and 1. Annotations that only
            have predicted labels with a probability lower than this value will be
            filtered out.
        :return: new Prediction object containing only annotations with a predicted
            probability higher than the confidence_threshold
        """
        annotations: List[Annotation] = []
        for annotation in self.annotations:
            max_prob = max([label.probability for label in annotation.labels])
            if max_prob > confidence_threshold:
                annotations.append(annotation)

        return Prediction(
            annotations=annotations,
            media_identifier=self.media_identifier,
            modified=datetime.now().isoformat(),
        )



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `annotation_readers` package contains the
:py:class:`~geti_sdk.annotation_readers.base_annotation_reader.AnnotationReader`
base class, which provides an interface for implementing custom annotation readers.

Annotation readers server to load annotation files in custom formats and convert them
to Intel® Geti™ format, such that they can be uploaded to an Intel® Geti™ project.

Module contents
---------------

.. automodule:: geti_sdk.annotation_readers.base_annotation_reader
   :members:
   :undoc-members:
   :show-inheritance:

.. automodule:: geti_sdk.annotation_readers.geti_annotation_reader
   :members:
   :undoc-members:
   :show-inheritance:

.. automodule:: geti_sdk.annotation_readers.datumaro_annotation_reader.datumaro_annotation_reader
   :members:
   :undoc-members:
   :show-inheritance:

.. automodule:: geti_sdk.annotation_readers.directory_tree_annotation_reader
   :members:
   :undoc-members:
   :show-inheritance:
"""

from .base_annotation_reader import AnnotationReader
from .datumaro_annotation_reader import DatumAnnotationReader
from .directory_tree_annotation_reader import DirectoryTreeAnnotationReader
from .geti_annotation_reader import GetiAnnotationReader

__all__ = [
    "AnnotationReader",
    "DatumAnnotationReader",
    "GetiAnnotationReader",
    "DirectoryTreeAnnotationReader",
]



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from typing import Any, ClassVar, Dict, List, Optional, Union

import attr

from geti_sdk.data_models import Algorithm
from geti_sdk.data_models.configurable_parameter_group import (
    PARAMETER_TYPES,
    ParameterGroup,
)
from geti_sdk.data_models.configuration_identifiers import (
    ComponentEntityIdentifier,
    HyperParameterGroupIdentifier,
)
from geti_sdk.data_models.utils import attr_value_serializer, deidentify


@attr.define
class ConfigurableParameters(ParameterGroup):
    """
    Representation of configurable parameters in GETi, as returned by the
    /configuration endpoint

    :var entity_identifier: Identification information for the entity to which the
        configurable parameters apply
    :var id: Unique database ID of the configurable parameters
    """

    _identifier_fields: ClassVar[str] = ["id"]

    entity_identifier: Union[
        HyperParameterGroupIdentifier, ComponentEntityIdentifier
    ] = attr.field(kw_only=True)
    id: Optional[str] = attr.field(default=None, kw_only=True)

    def deidentify(self) -> None:
        """
        Remove all unique database ID's from the configurable parameters
        """
        super().deidentify()
        deidentify(self)
        deidentify(self.entity_identifier)


@attr.define(slots=False)
class Configuration:
    """
    Representation of a set of configurable parameters in GETi, that apply to a project
    or task.
    """

    _identifier_fields: ClassVar[List[str]] = []

    components: List[ConfigurableParameters]

    def __attrs_post_init__(self):
        """
        Set configurable parameters as Configuration attributes
        """
        for parameter_name in self.get_all_parameter_names():
            setattr(self, parameter_name, self.get_parameter_by_name(parameter_name))

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the Configuration.
        """
        return attr.asdict(self, recurse=True, value_serializer=attr_value_serializer)

    def deidentify(self) -> None:
        """
        Remove all unique database ID's from the Configuration
        """
        deidentify(self)
        for config in self.components:
            config.deidentify()

    def __iter__(self):
        """
        Iterate over all parameters in the configuration.
        """
        parameter_names = self.get_all_parameter_names()
        for parameter_name in parameter_names:
            yield self.get_parameter_by_name(parameter_name)

    def get_all_parameter_names(self) -> List[str]:
        """
        Return a list of names of all configurable parameters within the task
        configuration.
        """
        parameters: List[str] = []
        for config in self.components:
            parameters.extend(config.parameter_names())
        return parameters

    def _set_parameter_value(
        self,
        parameter_name: str,
        value: Union[bool, float, int, str],
        group_name: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Prepare a dictionary that can be used for setting a parameter value in GETi.

        :param parameter_name: Name of the parameter to set
        :param value: Value to set for the parameter
        :param group_name: Optional name of the parameter group in which the parameter
            lives. If left as None (the default), this method will attempt to find the
            group automatically
        :return: Dictionary that can be passed directly to the
            /configuration POST endpoints to set the parameter
            value
        """
        if parameter_name not in self.get_all_parameter_names():
            raise ValueError(
                f"Parameter named '{parameter_name}' was not found in the configuration"
                f" {self}. Unable to prepare data to set "
                f"parameter value"
            )
        result: Dict[str, Any] = {}
        for config in self.components:
            parameter = config.get_parameter_by_name(parameter_name, group_name)
            if parameter is not None:
                result.update({"entity_identifier": config.entity_identifier.to_dict()})
                parameter_value_dict = {"name": parameter_name, "value": value}
                parameter.value = value
                if config.groups:
                    group = config.get_group_containing(parameter_name)
                    if group is not None:
                        result.update(
                            {
                                "groups": [
                                    {
                                        "name": group.name,
                                        "parameters": [parameter_value_dict],
                                    }
                                ]
                            }
                        )
                        break
                result.update({"parameters": [parameter_value_dict]})
                break
        return result

    def get_parameter_by_name(self, name: str) -> Optional[PARAMETER_TYPES]:
        """
        Return the configurable parameter named `name`. If no parameter by that name
        is found within the Configuration, this method returns None

        :param name: Name of the parameter to get
        :return: ConfigurableParameter object
        """
        parameters = [config.get_parameter_by_name(name) for config in self.components]
        parameters = [parameter for parameter in parameters if parameter is not None]
        if len(parameters) == 0:
            return None
        elif len(parameters) > 1:
            raise ValueError(
                f"Multiple parameters named '{name}' were found in the configuration "
                f"{self}. Unable to unambiguously retrieve "
                f"parameter."
            )
        return parameters[0]

    @property
    def component_configurations(self) -> List[ConfigurableParameters]:
        """
        Return all configurable parameters that are component-related.

        :return:
        """
        return [
            config
            for config in self.components
            if isinstance(config.entity_identifier, ComponentEntityIdentifier)
        ]

    def get_component_configuration(
        self, component: str
    ) -> Optional[ConfigurableParameters]:
        """
        Return the configurable parameters for a certain component. If
        no configuration is found for the specified component, this method returns None

        :param component: Name of the component to get the configuration for
        :return: ConfigurableParameters for the component
        """
        return next(
            (
                config
                for config in self.component_configurations
                if config.entity_identifier.component == component
            ),
            None,
        )


@attr.define(slots=False)
class GlobalConfiguration(Configuration):
    """
    Representation of the project-wide configurable parameters for a project in GETi.
    """

    def set_parameter_value(
        self,
        parameter_name: str,
        value: Union[bool, float, int, str],
        group_name: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        Prepare a dictionary that can be used for setting a parameter value in GETi.

        :param parameter_name: Name of the parameter to set
        :param value: Value to set for the parameter
        :param group_name: Optional name of the parameter group in which the parameter
            lives. If left as None (the default), this method will attempt to find the
            group automatically
        :return: Dictionary that can be passed directly to the
            /configuration POST endpoints to set the parameter
            value
        """
        result = self._set_parameter_value(parameter_name, value, group_name)
        return [result]

    def apply_identifiers(self, workspace_id: str, project_id: str):
        """
        Apply the unique database identifiers passed in `workspace_id`
        and `project_id` to all configurable parameters in the GlobalConfiguration.

        :param workspace_id: Workspace ID to assign
        :param project_id: Project ID to assign
        :return:
        """
        for config in self.components:
            config.entity_identifier.workspace_id = workspace_id
            config.entity_identifier.project_id = project_id

    @property
    def summary(self) -> str:
        """
        Return a string containing a very brief summary of the GlobalConfiguration.

        :return: string holding a very short summary of the GlobalConfiguration
        """
        summary_str = "Configuration for global components:\n"
        for configurable_parameters in self.components:
            summary_str += f"  {configurable_parameters.summary}\n"
        return summary_str


@attr.define(slots=False)
class TaskConfiguration(Configuration):
    """
    Representation of the configurable parameters for a task in GETi.
    """

    _identifier_fields: ClassVar[List[str]] = ["task_id"]

    task_id: Optional[str] = attr.field(default=None, kw_only=True)
    task_title: str = attr.field(kw_only=True)

    @property
    def model_configurations(self) -> List[ConfigurableParameters]:
        """
        Return all configurable parameters that are model-related.

        :return: List of configurable parameters
        """
        return [
            config
            for config in self.components
            if isinstance(config.entity_identifier, HyperParameterGroupIdentifier)
        ]

    def set_parameter_value(
        self,
        parameter_name: str,
        value: Union[bool, float, int, str],
        group_name: Optional[str] = None,
    ) -> Dict[str, Any]:
        """
        Prepare a dictionary that can be used for setting a configurable parameter
        value in GETi.

        :param parameter_name: Name of the parameter to set
        :param value: Value to set for the parameter
        :param group_name: Optional name of the parameter group in which the parameter
            lives. If left as None (the default), this method will attempt to find the
            group automatically
        :return: Dictionary that can be passed directly to the
            /configuration POST endpoints to set the parameter
            value
        """
        result = self._set_parameter_value(parameter_name, value, group_name)
        return {"components": [result]}

    def resolve_algorithm(self, algorithm: Algorithm):
        """
        Resolve the algorithm name and id of the model template for all hyper
        parameter groups in the task configuration.

        :param algorithm: Algorithm instance to which the hyper parameters belong
        :return:
        """
        for config in self.model_configurations:
            config.entity_identifier.resolve_algorithm(algorithm=algorithm)

    def apply_identifiers(
        self, workspace_id: str, project_id: str, task_id: str, model_storage_id: str
    ):
        """
        Apply the unique database identifiers passed in `workspace_id`,
        `project_id`, `task_id` and `model_storage_id` to all configurable parameters
        in the TaskConfiguration.

        :param workspace_id: Workspace ID to assign
        :param project_id: Project ID to assign
        :param task_id: Task ID to assign
        :param model_storage_id: Model storage ID to assign
        :return:
        """
        self.task_id = task_id
        for config in self.component_configurations:
            config.entity_identifier.workspace_id = workspace_id
            config.entity_identifier.project_id = project_id
            config.entity_identifier.task_id = task_id
        for config in self.model_configurations:
            config.entity_identifier.workspace_id = workspace_id
            config.entity_identifier.model_storage_id = model_storage_id

    @property
    def summary(self) -> str:
        """
        Return a string containing a very brief summary of the TaskConfiguration.

        :return: string holding a very short summary of the TaskConfiguration
        """
        summary_str = f"Configuration for {self.task_title}:\n"
        for configurable_parameters in self.components:
            summary_str += f"  {configurable_parameters.summary}\n"
        return summary_str


@attr.define
class FullConfiguration:
    """
    Representation of the full configuration (both global and task-chain) for a
    project in GETi.
    """

    global_: GlobalConfiguration
    task_chain: List[TaskConfiguration]

    def deidentify(self) -> None:
        """
        Remove all unique database ID's from the Configuration.
        """
        self.global_.deidentify()
        for task_config in self.task_chain:
            task_config.deidentify()

    def to_dict(self) -> Dict[str, Any]:
        """
        Return the dictionary representation of the FullConfiguration.
        """
        result = attr.asdict(self, recurse=True, value_serializer=attr_value_serializer)
        global_dict = result.pop("global_")
        result.update({"global": global_dict["components"]})
        return result

    @property
    def summary(self) -> str:
        """
        Return a string containing a very brief summary of the FullConfiguration.

        :return: string holding a very short summary of the FullConfiguration
        """
        summary_str = "Full project configuration:\n"
        summary_str += f"  {self.global_.summary}\n"
        for task_configuration in self.task_chain:
            summary_str += f"  {task_configuration.summary}\n"
        return summary_str



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.
from typing import List, Optional

import numpy as np

from geti_sdk.data_models import Prediction
from geti_sdk.deployment.inference_hook_interfaces import PostInferenceTrigger

SUPPORTED_MODES = ["greater", "lower", "equal"]


class ObjectCountTrigger(PostInferenceTrigger):
    """
    Post inference trigger based on the number of predicted objects in an image.

    The operation mode of the trigger can be specified: It will activate either if the
    object count is above the threshold, below the threshold or exactly equal to the
    threshold value. This behaviour can be controlled via the `mode` parameter.

    :param threshold: Threshold value for the total number of objects in a prediction
    :param label_names: Optional list of classes/labels to account for in the object
        counting. If specified, only objects with those labels will count towards the
        total
    :param mode: Operation mode for the trigger, which defines when the trigger will
        activate. Options are:

            - `greater`  -> Trigger activates if object count is greater than threshold
            - `lower`    -> Trigger activates if object count is lower than threshold
            - `equal`    -> Trigger activates if object count is equal to threshold

        Defaults to `greater`
    """

    def __init__(
        self,
        threshold: int = 1,
        label_names: Optional[List[str]] = None,
        mode: str = "greater",
    ):
        super().__init__(threshold=threshold)
        self.label_names = label_names
        self.filter_labels = False

        if label_names is not None:
            self.filter_labels = True
            self._repr_info_ += f", label_names={label_names}"

        lower_mode = mode.lower()
        if lower_mode not in SUPPORTED_MODES:
            raise ValueError(
                f"Invalid mode `{mode}`. Valid options are: {SUPPORTED_MODES}"
            )
        self.mode = lower_mode
        self._repr_info_ += f", mode={lower_mode}"

    def __call__(self, image: np.ndarray, prediction: Prediction) -> float:
        """
        Compute a trigger score for the `image` and corresponding `prediction`.

        :param image: Numpy array representing an image
        :param prediction: Prediction object corresponding to the inference result
            for the image.
        :return: Float representing the score for the input
        """
        if not self.filter_labels:
            n_objects = len(prediction.annotations)
            if n_objects != 1:
                return n_objects
            else:
                # Prediction might be a 'No object', we have to check for this
                predicted_labels = prediction.annotations[0].labels
                if len(predicted_labels) > 1:
                    return 1.0
                else:
                    if predicted_labels[0].name.lower() == "no object":
                        return 0.0
                    else:
                        return 1.0
        else:
            object_count: int = 0
            for predicted_object in prediction.annotations:
                for label in predicted_object.labels:
                    if label.name in self.label_names:
                        object_count += 1
        return object_count

    def get_decision(self, score: float) -> bool:
        """
        Make a decision based on a previously computed `score` and the threshold defined
        for the trigger

        :param score: Float representing the score for a certain image, prediction pair
        :return: Boolean indicating whether the trigger conditions are met (True), or
            not (False)
        """
        if self.mode == "lower":
            return score < self.threshold
        elif self.mode == "greater":
            return score > self.threshold
        elif self.mode == "equal":
            return score == self.threshold



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
import re
from importlib import resources
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple

import numpy as np
from pathvalidate import sanitize_filepath

from geti_sdk.data_models import (
    Annotation,
    Label,
    OptimizedModel,
    Prediction,
    Project,
    ScoredLabel,
    Task,
    TaskType,
)
from geti_sdk.data_models.shapes import Rectangle
from geti_sdk.deployment.data_models import ROI, IntermediateInferenceResult

try:
    OVMS_README_PATH = str(
        resources.files("geti_sdk.deployment.resources") / "OVMS_README.md"
    )
except AttributeError:
    with resources.path("geti_sdk.deployment.resources", "OVMS_README.md") as data_path:
        OVMS_README_PATH = str(data_path)


def generate_ovms_model_name(
    project: Project, model: OptimizedModel, omit_version: bool = True
) -> str:
    """
    Generate a valid name for a model to be uploaded to the OpenVINO Model Server.

    :param project: Project the model belongs to
    :param model: The model for which to generate the name
    :param omit_version: True not to include the version of the model in the name.
        This is useful if the name needs to be used to create a directory structure,
        since including the version will break this.
    :return: String containing the name of the model
    """
    model_name = sanitize_filepath(project.name + "_" + model.name)
    model_name = model_name.replace(" ", "_")
    model_name = model_name.replace("-", "_")
    model_name = model_name.lower()
    if model.version is not None and not omit_version:
        model_name += f":{model.version}"
    return model_name


def generate_ovms_model_address(ovms_address: str, model_name: str) -> str:
    """
    Generate a string representing a valid model address for the OpenVINO Model Server,
    from a given `ovms_address` and `model_name`.

    The format expected for model addresses by OVMS is:

        <address>:<port>/models/<model_name>[:<model_version>]

    Where `address` must not contain the protocol (http or https)

    :param ovms_address: IP address or URL pointing to the OVMS instance, including
        port specification
    :param model_name: Name of the model
    :return: String containing the model address
    """
    model_address = f"{ovms_address}/models/{model_name}"
    if model_address.startswith("https://"):
        model_address = model_address[8:]
    if model_address.startswith("http://"):
        model_address = model_address[7:]
    return model_address


def target_device_is_ovms(device: str) -> bool:
    """
    Return True if the target `device` specified is a URL or IP address, False otherwise

    :param device: Target device string to check
    :return: True if the string represents a URL or IP address, False otherwise
    """
    # Check if 'device' has been specified as a URL or IP address.
    server_pattern = re.compile(
        r"^((https?://)|(www.))(?:([a-zA-Z]+)|(\d+\.\d+\.\d+\.\d+)):\d{1,5}?$"
    )
    return server_pattern.match(device) is not None


def rgb_to_hex(rgb: Tuple[int, int, int]) -> str:
    """
    Convert an RGB color value to its corresponding hexadecimal representation.

    :param rgb: A tuple representing the RGB color value, where each element is an integer between 0 and 255.
    :return: The hexadecimal representation of the RGB color value.

    _Example:

        >>> rgb_to_hex((255, 0, 0))
        '#ff0000'
    """
    return "#{:02x}{:02x}{:02x}".format(rgb[0], rgb[1], rgb[2])


def get_package_version_from_requirements(
    requirements_path: os.PathLike, package_name: str
) -> str:
    """
    Get the version of a package from a requirements file.

    :param requirements_path: The requirements file path
    :param package_name: The name of the package to get the version of
    :return: The version of the package as a string, empty line if the package is not found.
    :raises: ValueError If the requirements file is not found.
    """
    if not package_name:
        return ""

    requirements_path = Path(requirements_path).resolve()
    if not requirements_path.exists():
        raise ValueError(f"Requirements file {requirements_path} not found")

    for line in requirements_path.read_text().split("\n"):
        # The otx may be installed from a GitHub url
        if line.startswith(package_name) and "==" in line:
            return line.split("==")[1].strip()
    return ""


def assign_empty_label(
    prediction: Prediction, image: np.ndarray, empty_label: Optional[Label] = None
):
    """
    Modify the `prediction` in place to assign an empty label, if the prediction
    does not contain any objects.

    :param prediction: Prediction object to assign the empty label to, if needed
    :param image: The original image for which the prediction was generated
    :param empty_label: The Label object representing the empty label to assign
    """
    # Empty label is not generated by prediction postprocessing correctly, append
    # it here if there are no other predictions
    height, width = image.shape[0:2]
    if len(prediction.annotations) == 0:
        if empty_label is not None:
            prediction.append(
                Annotation(
                    shape=Rectangle(x=0, y=0, width=width, height=height),
                    labels=[ScoredLabel.from_label(empty_label, probability=1)],
                )
            )


def flow_control(
    intermediate_result: IntermediateInferenceResult, task: Task
) -> Tuple[List[ROI], List[np.ndarray]]:
    """
    Logic for the flow control tasks

    :param intermediate_result: IntermediateInferenceResult object, containing the
        predictions so far, as well as the rois.
    :param task: The task to perform the flow control operation for
    :return: Tuple containing:
        - rois -> A list of rois generated by the flow control task, to be passed
            down to the next task in the task chain
        - image_views -> A list of numpy arrays containing the image data in the
            corresponding roi
    """
    # CROP task
    if task.type == TaskType.CROP:
        rois = intermediate_result.filter_rois(label=None)
        image_views = intermediate_result.generate_views(rois)
    else:
        raise NotImplementedError(
            f"Unable to run inference for the pipeline in the deployed "
            f"project: Unsupported task type {task.type} found."
        )
    return rois, image_views


def pipeline_callback_factory(
    is_first_task: bool,
    is_final_task: bool,
    add_flow_control: bool,
    current_task: Task,
    deployment,
    final_callback: Callable[[Prediction, Optional[Any]], None],
    next_trainable_task: Optional[Task] = None,
    next_task: Optional[Task] = None,
) -> Callable[[Prediction, Any], None]:
    """
    Generate callback functions for pipeline inference
    """
    if is_first_task:

        def callback(result: Prediction, runtime_data: Tuple[np.ndarray, Any]):
            """
            Process inference results for the first task in a task chain, this
            initializes the intermediate result
            """
            image, runtime_user_data = runtime_data
            rois = [
                ROI.from_annotation(annotation) for annotation in result.annotations
            ]

            if len(rois) == 0:
                # Empty prediction. Assign empty label and run final callback immediately
                assign_empty_label(result, image, deployment._empty_label)
                final_callback(image, result, runtime_user_data)
                return

            intermediate_result = IntermediateInferenceResult(
                image=image, prediction=result, rois=rois
            )

            if add_flow_control:
                rois, image_views = flow_control(intermediate_result, next_task)
                intermediate_result.rois = rois
            else:
                image_views = intermediate_result.generate_views()
            for roi, view in zip(rois, image_views):
                deployment._infer_task_async(
                    image=view,
                    explain=False,
                    task=next_trainable_task,
                    runtime_data=(image, roi, intermediate_result, runtime_user_data),
                )

    else:

        def callback(
            result: Prediction,
            runtime_data: Tuple[np.ndarray, ROI, IntermediateInferenceResult, Any],
        ):
            """
            Process inference results for subsequent tasks
            """
            image, roi, intermediate_result, runtime_user_data = runtime_data

            if current_task.is_global:
                # Global tasks add their labels to the existing shape in the ROI
                intermediate_result.extend_annotations(result.annotations, roi=roi)
            else:
                # Local tasks create new shapes in the image coordinate system,
                # and generate ROI's corresponding to the new shapes
                for annotation in result.annotations:
                    intermediate_result.append_annotation(annotation, roi=roi)
                    new_roi = ROI.from_annotation(annotation)
                    intermediate_result.add_to_infer_queue(
                        new_roi.to_absolute_coordinates(parent_roi=roi)
                    )
            intermediate_result.increment_infer_counter()
            if intermediate_result.all_rois_inferred():
                # In this case, all ROIS from the preceding task have been inferred.
                # The infer queue now becomes the new ROIs for the intermediate result
                new_rois = intermediate_result.get_infer_queue()
                intermediate_result.reset_infer_counter()
                intermediate_result.rois = new_rois

                if add_flow_control:
                    rois, image_views = flow_control(intermediate_result, next_task)
                    intermediate_result.rois = rois
                else:
                    image_views = intermediate_result.generate_views()

                if not is_final_task:
                    for roi, view in zip(rois, image_views):
                        deployment._infer_task_async(
                            image=view,
                            explain=False,
                            task=next_trainable_task,
                            runtime_data=(
                                image,
                                roi,
                                intermediate_result,
                                runtime_user_data,
                            ),
                        )
                else:
                    prediction = intermediate_result.prediction
                    final_callback(image, prediction, runtime_user_data)

    return callback



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from enum import Enum

from geti_sdk.data_models.enums.task_type import TaskType


class Domain(Enum):
    """
    Enum representing the different task domains in Intel® Geti™ projects.
    """

    DETECTION = "DETECTION"
    SEGMENTATION = "SEGMENTATION"
    CLASSIFICATION = "CLASSIFICATION"
    ANOMALY_CLASSIFICATION = "ANOMALY_CLASSIFICATION"
    ANOMALY_DETECTION = "ANOMALY_DETECTION"
    ANOMALY_SEGMENTATION = "ANOMALY_SEGMENTATION"
    INSTANCE_SEGMENTATION = "INSTANCE_SEGMENTATION"
    ROTATED_DETECTION = "ROTATED_DETECTION"
    ANOMALY = "ANOMALY"

    def __str__(self) -> str:
        """
        Return the string representation of the Domain instance.
        """
        return self.value

    @classmethod
    def from_task_type(cls, task_type: TaskType) -> "Domain":
        """
        Return the Domain corresponding to a certain task type.

        :param task_type: TaskType to retrieve the domain for
        :return: Domain corresponding to task_type
        """
        return cls[task_type.name]



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

"""
Introduction
------------

The `prediction_visualization` package provides classes for visualizing models predictions.
Currently, the user interfaces to this package are available in the :py:mod:`~geti_sdk.utils.plot_helpers` module.

Module contents
---------------

.. automodule:: geti_sdk.prediction_visualization.visualizer
   :members:
   :undoc-members:
   :show-inheritance:

.. autoclass:: geti_sdk.prediction_visualization.shape_drawer.ShapeDrawer
   :members:
   :undoc-members:
   :show-inheritance:
"""

from .visualizer import Visualizer

__all__ = ["Visualizer"]



# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import logging
import time
from dataclasses import dataclass, field
from multiprocessing import Value
from queue import Empty, PriorityQueue
from typing import Any, Optional

import numpy as np

from geti_sdk.data_models import Prediction


@dataclass(order=True)
class IndexedResult:
    """
    Indexed container class that holds the result of model inference, with the
    original image and any additional data passed at runtime
    """

    index: int
    prediction: Prediction = field(compare=False)
    image: np.ndarray = field(compare=False)
    runtime_data: Optional[Any] = field(compare=False, default=None)


class OrderedResultBuffer:
    """
    A buffer for inference results, ordered according to an index
    """

    def __init__(self, minsize: int = 30, maxsize: Optional[int] = None):
        """
        Buffer inference results and assign an index, so that they can be retrieved in
        an ordered fashion.

        :param minsize: Minimal number of items to keep in the buffer. This is the
            minimal size the buffer should have before yielding items.
        :param maxsize: Maximum number of items that the buffer can contain. When the
            buffer contains this number of items, any attempt to add an item will be
            blocking.
        """
        if isinstance(maxsize, int):
            queue_args = {"maxsize": maxsize}
        else:
            queue_args = {}
        self._queue = PriorityQueue[IndexedResult](**queue_args)
        self._minsize = minsize
        self.__queue_args = queue_args
        self._put_count = Value("i", 0)
        self._get_count = Value("i", 0)

        if maxsize is not None and minsize >= maxsize:
            raise ValueError("minsize must be smaller than maxsize")

        logging.info(
            f"OrderedBuffer intialized with `minsize={minsize}` and "
            f"`maxsize={maxsize}`"
        )

    def put(
        self,
        index: int,
        image: np.ndarray,
        prediction: Prediction,
        runtime_data: Optional[Any] = None,
    ):
        """
        Add an image, prediction and corresponding runtime data as an entry to the
        buffer

        :param index: The index of the result item
        :param image: The original image for which the prediction was generated
        :param prediction: The prediction for the image
        :param runtime_data: Any additional data passed at runtime
        """
        queue_item = IndexedResult(
            index=index, image=image, prediction=prediction, runtime_data=runtime_data
        )
        self._queue.put(queue_item)
        with self._put_count.get_lock():
            self._put_count.value += 1

    def get(self, timeout: float = 0, empty_buffer: bool = False) -> IndexedResult:
        """
        Get the next item from the buffer

        :param timeout: Timeout in seconds
        :param empty_buffer: True to allow retrieving all items from the buffer, i.e.
            reducing the number of available buffered items to below the minsize.
        :return: An IndexedResult object, containing the index, the original image,
            the prediction and any additional data assigned at runtime
        """
        t_start = time.time()
        item: Optional[IndexedResult] = None
        if not empty_buffer:
            while timeout == 0 or time.time() - t_start < timeout:
                if self._queue.qsize() > self._minsize:
                    try:
                        item = self._queue.get(block=False)
                    except Empty:
                        time.sleep(1e-9)
                        continue
                    break
                time.sleep(1e-9)
        else:
            item = self._queue.get(block=True, timeout=timeout + 1e-9)
        if item is None:
            raise Empty
        self._queue.task_done()
        with self._get_count.get_lock():
            self._get_count.value += 1
        return item

    def reset(self) -> None:
        """
        Clear all remaining items from the buffer, and reset the index.
        """
        self._queue = PriorityQueue[IndexedResult](**self.__queue_args)
        with self._put_count.get_lock():
            self._put_count.value = 0
        with self._get_count.get_lock():
            self._get_count.value = 0

    @property
    def is_empty(self) -> bool:
        """
        Return True if the buffer is empty, False otherwise

        :return: True if the buffer is empty, False otherwise
        """
        return self._queue.qsize() == 0

    @property
    def total_items_buffered(self) -> int:
        """
        Return the total number of items that have been put to the buffer
        """
        with self._put_count.get_lock():
            return self._put_count.value

    @property
    def current_buffer_size(self) -> int:
        """
        Return the number of items that are currently in the buffer
        """
        with self._put_count.get_lock(), self._get_count.get_lock():
            return self._put_count.value - self._get_count.value()
